"""State definitions for the Room Agent LangGraph workflow."""

from __future__ import annotations

from typing import Any, Literal, NotRequired, TypedDict


class IntentResult(TypedDict, total=False):
    """Structured intent classification result."""

    name: str
    confidence: float


class ExecutionError(TypedDict, total=False):
    """Structured execution error payload."""

    type: str
    message: str
    source_node: str
    retryable: bool


class RoomAgentGraphState(TypedDict, total=False):
    """Shared graph state for the next-generation Room Agent workflow."""

    user_input: str
    conversation_text: str
    intent: IntentResult
    need_tool_call: bool
    candidate_tools: list[dict[str, Any]]
    selected_tools: list[dict[str, Any]]
    tool_call_history: list[dict[str, Any]]
    plan: dict[str, Any]
    human_review: dict[str, Any]
    execution_args: dict[str, Any]
    execution_result: dict[str, Any]
    error: ExecutionError
    metadata: dict[str, Any]
    artifacts: dict[str, Any]
    status: Literal["initialized", "completed", "failed"]
    next_action: NotRequired[str]



"""LLM provider factory using LangChain's OpenAI-compatible integration."""

from __future__ import annotations

from typing import Protocol

from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, SystemMessage
from langchain_openai import ChatOpenAI

from config.settings import LLMModelSettings, LLMRole, LLMSettings


class ChatProvider(Protocol):
    async def complete_text(
        self,
        messages: list[dict[str, str]],
        *,
        temperature: float = 0.2,
        json_mode: bool = False,
    ) -> str: ...


class ProviderError(RuntimeError):
    """Raised when an upstream LLM provider returns an invalid response."""


class OpenAICompatibleProvider:
    def __init__(self, settings: LLMModelSettings) -> None:
        self.settings = settings
        self.model = ChatOpenAI(
            model=settings.model,
            api_key=settings.api_key,
            base_url=settings.base_url,
            temperature=settings.temperature,
            extra_body={"enable_thinking": False},
        )

    async def complete_text(
        self,
        messages: list[dict[str, str]],
        *,
        temperature: float = 0.2,
        json_mode: bool = False,
    ) -> str:
        model = self.model.bind(temperature=temperature)
        if json_mode:
            model = model.bind(response_format={"type": "json_object"})

        response = await model.ainvoke(_to_langchain_messages(messages))
        content = _normalize_ai_message_content(response)
        if not content:
            raise ProviderError("OpenAI-compatible provider returned empty content.")
        return content


class LLMProviderRegistry:
    def __init__(
        self,
        *,
        powerful: ChatProvider | None,
        low_cost: ChatProvider | None,
    ) -> None:
        fallback = low_cost or powerful
        self._providers = {
            LLMRole.POWERFUL: powerful or fallback,
            LLMRole.LOW_COST: low_cost or fallback,
        }

    def get(self, role: LLMRole) -> ChatProvider | None:
        return self._providers.get(role)


def create_llm_provider(settings: LLMModelSettings) -> ChatProvider | None:
    if not settings.has_credentials:
        return None

    return OpenAICompatibleProvider(settings)

def create_llm_provider_registry(settings: LLMSettings) -> LLMProviderRegistry:
    return LLMProviderRegistry(
        powerful=create_llm_provider(settings.for_role(LLMRole.POWERFUL)),
        low_cost=create_llm_provider(settings.for_role(LLMRole.LOW_COST)),
    )


def _to_langchain_messages(messages: list[dict[str, str]]) -> list[BaseMessage]:
    converted: list[BaseMessage] = []
    for message in messages:
        role = message["role"]
        content = message["content"]
        if role == "system":
            converted.append(SystemMessage(content=content))
        elif role == "assistant":
            converted.append(AIMessage(content=content))
        else:
            converted.append(HumanMessage(content=content))
    return converted


def _normalize_ai_message_content(message: AIMessage) -> str:
    if isinstance(message.content, str):
        return message.content
    if isinstance(message.content, list):
        return "".join(
            block.get("text", "") if isinstance(block, dict) else str(block)
            for block in message.content
        )
    return str(message.content)



from __future__ import annotations

import asyncio
from pathlib import Path

import pytest
from openai import APIConnectionError

from app.server import initialize_runtime_dependencies
from config.settings import load_settings
from graph.entry import compile_graph
from integrations.llm_provider import create_llm_provider_registry


ROOT = Path(__file__).resolve().parents[1]
ROOM_AGENT_CONFIG = ROOT / "config" / "examples" / "room_agent.example.yaml"
LLM_CONFIG = Path(__file__).resolve().parent / "fixtures" / "llm.yaml"


def _require_llm_fixture() -> None:
    if not LLM_CONFIG.exists():
        pytest.skip(
            "Missing room-agent/tests/fixtures/llm.yaml. "
            "Create it locally from room-agent/config/examples/llm.example.yaml "
            "and fill in real provider credentials for the low_cost and powerful roles."
        )


def _initialize_real_runtime() -> None:
    _require_llm_fixture()
    settings = load_settings(
        config_path=str(ROOM_AGENT_CONFIG),
        llm_config_path=str(LLM_CONFIG),
    )
    registry = create_llm_provider_registry(settings.llm)
    initialize_runtime_dependencies(
        settings=settings,
        llm_provider_registry=registry,
    )


def _invoke_graph_or_skip(user_input: str) -> dict[str, object]:
    try:
        return asyncio.run(compile_graph().ainvoke({"user_input": user_input}))
    except APIConnectionError as exc:
        pytest.skip(f"Skipping smoke test because upstream LLM is unreachable: {exc}")


def test_graph_routes_chat_to_direct_response():
    _initialize_real_runtime()

    final_state = _invoke_graph_or_skip("你好")

    assert final_state["need_tool_call"] is False
    assert final_state["next_action"] == "direct_response"
    assert final_state["execution_result"]["type"] == "text"
    assert final_state["execution_result"]["message"]


def test_graph_routes_tool_request_to_tool_selection():
    _initialize_real_runtime()

    final_state = _invoke_graph_or_skip("帮我打开卧室的灯")

    assert final_state["need_tool_call"] is True
    assert final_state["next_action"] == "agent_execution"
    assert "message" in final_state["execution_result"]



from __future__ import annotations

import asyncio

from app.server import MCPHealthStatus, get_mcp_health_status, initialize_runtime_dependencies, probe_home_assistant_mcp
from config.settings import AgentSettings, HomeAssistantMCPSettings, RuntimeSettings, Settings


class PromptResult:
    def __init__(self, prompts):
        self.prompts = prompts


class HealthyClient:
    async def list_prompts(self, server_name: str):
        assert server_name == "home_assistant"
        return PromptResult(prompts=["device_control"])


class FailingClient:
    async def list_prompts(self, server_name: str):
        raise RuntimeError(f"server unavailable: {server_name}")


class GroupedFailingClient:
    async def list_prompts(self, server_name: str):
        raise ExceptionGroup(
            "unhandled errors in a TaskGroup",
            [PermissionError(f"401 Unauthorized for {server_name}")],
        )


def test_probe_home_assistant_mcp_marks_healthy_from_prompt_listing() -> None:
    status = asyncio.run(
        probe_home_assistant_mcp(
            HealthyClient(),
            HomeAssistantMCPSettings(
                enabled=True,
                server_name="home_assistant",
                base_url="http://ha.local:8123",
            ),
        )
    )

    assert status.enabled is True
    assert status.healthy is True
    assert status.prompt_count == 1
    assert status.error is None


def test_probe_home_assistant_mcp_failure_does_not_raise() -> None:
    status = asyncio.run(
        probe_home_assistant_mcp(
            FailingClient(),
            HomeAssistantMCPSettings(
                enabled=True,
                server_name="home_assistant",
                base_url="http://ha.local:8123",
            ),
        )
    )

    assert status.enabled is True
    assert status.healthy is False
    assert status.prompt_count is None
    assert status.error_type == "RuntimeError"
    assert status.error == "RuntimeError: server unavailable: home_assistant"


def test_probe_home_assistant_mcp_expands_exception_group_details() -> None:
    status = asyncio.run(
        probe_home_assistant_mcp(
            GroupedFailingClient(),
            HomeAssistantMCPSettings(
                enabled=True,
                server_name="home_assistant",
                base_url="http://ha.local:8123",
            ),
        )
    )

    assert status.enabled is True
    assert status.healthy is False
    assert status.error_type == "ExceptionGroup"
    assert "PermissionError: 401 Unauthorized for home_assistant" in status.error


def test_initialize_runtime_dependencies_exposes_health_status_snapshot() -> None:
    initialize_runtime_dependencies(
        settings=Settings.model_construct(
            agent=AgentSettings(),
            llm=None,
            runtime=RuntimeSettings(
                room_agent_config_path="room-agent/config/examples/room_agent.example.yaml"
            ),
        ),
        llm_provider_registry=object(),
        mcp_client=None,
        mcp_health_status=MCPHealthStatus(
            enabled=True,
            healthy=False,
            server_name="home_assistant",
            error_type="RuntimeError",
            error="server unavailable",
        ),
    )

    health = get_mcp_health_status()

    assert health["enabled"] is True
    assert health["healthy"] is False
    assert health["server_name"] == "home_assistant"
    assert health["error_type"] == "RuntimeError"
    assert health["error"] == "server unavailable"



from __future__ import annotations

import asyncio
import importlib
import json

from graph.entry import initialize_request
from graph.nodes import direct_response as direct_response_function
from graph.nodes import intent_recognition as intent_recognition_function
from graph.nodes import tool_selection as tool_selection_function
from graph.subgraphs.agent_execution import agent_execution as agent_execution_function


direct_response_module = importlib.import_module("graph.nodes.direct_response")
intent_recognition_module = importlib.import_module("graph.nodes.intent_recognition")
tool_selection_module = importlib.import_module("graph.nodes.tool_selection")
agent_execution_module = importlib.import_module("graph.subgraphs.agent_execution.entry")


class RecordingProvider:
    def __init__(self, response: str) -> None:
        self.response = response
        self.calls: list[dict[str, object]] = []

    async def complete_text(self, messages, *, temperature=0.2, json_mode=False):
        self.calls.append(
            {
                "messages": messages,
                "temperature": temperature,
                "json_mode": json_mode,
            }
        )
        return self.response


class SequenceProvider:
    def __init__(self, responses: list[str]) -> None:
        self.responses = responses
        self.calls: list[dict[str, object]] = []

    async def complete_text(self, messages, *, temperature=0.2, json_mode=False):
        self.calls.append(
            {
                "messages": messages,
                "temperature": temperature,
                "json_mode": json_mode,
            }
        )
        if not self.responses:
            raise AssertionError("No more mock responses configured.")
        return self.responses.pop(0)


class FakeToolDescriptor:
    def __init__(self, name: str, description: str, args_schema: dict[str, object]) -> None:
        self.name = name
        self.description = description
        self.args_schema = args_schema
        self.args = args_schema

    def model_dump(self) -> dict[str, object]:
        return {
            "name": self.name,
            "description": self.description,
            "args_schema": self.args_schema,
        }


class FakeMCPClient:
    def __init__(self, descriptors: list[FakeToolDescriptor]) -> None:
        self.descriptors = descriptors

    async def get_tools(self) -> list[FakeToolDescriptor]:
        return self.descriptors


class FakeInvokableTool(FakeToolDescriptor):
    def __init__(
        self,
        name: str,
        description: str,
        args_schema: dict[str, object],
        *,
        result: object | None = None,
        error: Exception | None = None,
    ) -> None:
        super().__init__(name, description, args_schema)
        self.result = result
        self.error = error
        self.calls: list[dict[str, object]] = []

    async def ainvoke(self, args: dict[str, object]) -> object:
        self.calls.append(args)
        if self.error is not None:
            raise self.error
        return self.result


def test_initialize_request_uses_user_input_as_fallback_conversation_text() -> None:
    result = initialize_request({"user_input": "你好"})

    assert result["user_input"] == "你好"
    assert result["conversation_text"] == "你好"


def test_direct_response_prefers_conversation_text(monkeypatch) -> None:
    provider = RecordingProvider("简短回复")
    monkeypatch.setattr(direct_response_module, "_get_low_cost_provider", lambda: provider)

    state = {
        "user_input": "最后一句",
        "conversation_text": "user: 第一轮\nassistant: 第二轮\n\nCurrent user input:\n最后一句",
        "intent": {"name": "chat"},
    }
    result = asyncio.run(direct_response_function(state))

    assert result["execution_result"]["message"] == "简短回复"
    assert provider.calls[0]["messages"][1]["content"] == state["conversation_text"]


def test_intent_recognition_prefers_conversation_text(monkeypatch) -> None:
    provider = RecordingProvider('{"intent_name":"chat","need_tool_call":false}')
    monkeypatch.setattr(intent_recognition_module, "_get_low_cost_provider", lambda: provider)

    class FakeParser:
        def __init__(self, llm_provider):
            assert llm_provider is provider

        async def __call__(self, raw_output, schema):
            assert raw_output
            assert schema["required"] == ["intent_name", "need_tool_call"]
            return {"intent_name": "chat", "need_tool_call": False}

    monkeypatch.setattr(intent_recognition_module, "JsonParserWithRepair", FakeParser)

    state = {
        "user_input": "最后一句",
        "conversation_text": "user: 你好\nassistant: 在的\n\nCurrent user input:\n最后一句",
    }
    result = asyncio.run(intent_recognition_function(state))

    assert result["next_action"] == "direct_response"
    assert state["conversation_text"] in provider.calls[0]["messages"][1]["content"]


def test_tool_selection_prefers_conversation_text_and_selects_tools(monkeypatch) -> None:
    provider = RecordingProvider(
        '{"selected_tool_names":["weather_lookup","calendar_lookup"],"comment":"weather and schedule"}'
    )
    monkeypatch.setattr(tool_selection_module, "_get_low_cost_provider", lambda: provider)
    monkeypatch.setattr(
        tool_selection_module,
        "_get_mcp_client",
        lambda: FakeMCPClient(
            [
                FakeToolDescriptor("weather_lookup", "查询天气", {"query": {"type": "string"}}),
                FakeToolDescriptor("calendar_lookup", "查询日历", {"query": {"type": "string"}}),
                FakeToolDescriptor("news_lookup", "查询新闻", {"query": {"type": "string"}}),
            ]
        ),
    )

    state = {
        "user_input": "最后一句",
        "conversation_text": "user: 帮我看看今天安排和天气\nassistant: 好的\n\nCurrent user input:\n最后一句",
        "intent": {"name": "information_query"},
    }
    result = asyncio.run(tool_selection_function(state))

    assert provider.calls[0]["json_mode"] is True
    assert provider.calls[0]["temperature"] == 0
    payload = json.loads(provider.calls[0]["messages"][1]["content"])
    assert payload["user_input"] == state["conversation_text"]
    assert len(result["candidate_tools"]) == 3
    assert [tool["name"] for tool in result["selected_tools"]] == [
        "weather_lookup",
        "calendar_lookup",
    ]
    assert result["execution_result"]["type"] == "tool_selection"
    assert result["execution_result"]["comment"] == "weather and schedule"


def test_tool_selection_logs_to_stderr_when_llm_returns_zero_tools(monkeypatch, capsys) -> None:
    provider = RecordingProvider('{"selected_tool_names":[],"comment":"no relevant tool"}')
    monkeypatch.setattr(tool_selection_module, "_get_low_cost_provider", lambda: provider)
    monkeypatch.setattr(
        tool_selection_module,
        "_get_mcp_client",
        lambda: FakeMCPClient(
            [FakeToolDescriptor("weather_lookup", "查询天气", {"query": {"type": "string"}})]
        ),
    )

    result = asyncio.run(
        tool_selection_function({"user_input": "随便聊聊", "intent": {"name": "chat"}})
    )
    captured = capsys.readouterr()

    assert result["selected_tools"] == []
    assert result["execution_result"]["comment"] == "no relevant tool"
    assert "no tools selected" in captured.err
    assert "no relevant tool" in captured.err


def test_tool_selection_filters_unknown_names_and_deduplicates(monkeypatch, capsys) -> None:
    provider = RecordingProvider(
        '{"selected_tool_names":["ghost_tool","weather_lookup","weather_lookup"],"comment":"best effort"}'
    )
    monkeypatch.setattr(tool_selection_module, "_get_low_cost_provider", lambda: provider)
    monkeypatch.setattr(
        tool_selection_module,
        "_get_mcp_client",
        lambda: FakeMCPClient(
            [FakeToolDescriptor("weather_lookup", "查询天气", {"query": {"type": "string"}})]
        ),
    )

    result = asyncio.run(
        tool_selection_function({"user_input": "查天气", "intent": {"name": "info"}})
    )
    captured = capsys.readouterr()

    assert [tool["name"] for tool in result["selected_tools"]] == ["weather_lookup"]
    assert captured.err == ""


def test_tool_selection_logs_when_no_candidate_tools(monkeypatch, capsys) -> None:
    provider = RecordingProvider('{"selected_tool_names":["weather_lookup"],"comment":"unused"}')
    monkeypatch.setattr(tool_selection_module, "_get_low_cost_provider", lambda: provider)
    monkeypatch.setattr(tool_selection_module, "_get_mcp_client", lambda: FakeMCPClient([]))

    result = asyncio.run(
        tool_selection_function({"user_input": "帮我开灯", "intent": {"name": "device_control"}})
    )
    captured = capsys.readouterr()

    assert result["candidate_tools"] == []
    assert result["selected_tools"] == []
    assert result["execution_result"]["comment"] == "No MCP tools are available for this request."
    assert provider.calls == []
    assert "no tools selected" in captured.err


def test_agent_execution_returns_structured_final_output(monkeypatch) -> None:
    provider = SequenceProvider(
        [
            (
                '{"step_type":"final_output","is_done":true,'
                '"final_output":{"message":"已为你生成结果","summary":"done","metadata":{"source":"planner"}}}'
            )
        ]
    )
    monkeypatch.setattr(agent_execution_module, "_get_powerful_provider", lambda: provider)
    monkeypatch.setattr(agent_execution_module, "get_mcp_client", lambda: None)

    result = asyncio.run(
        agent_execution_function(
            {
                "user_input": "帮我开灯",
                "conversation_text": "user: 帮我开灯",
                "intent": {"name": "device_control"},
                "selected_tools": [{"name": "light_control", "description": "控制灯光"}],
            }
        )
    )

    assert result["status"] == "completed"
    assert result["next_action"] == "agent_execution"
    assert result["execution_result"]["type"] == "agent_final_output"
    assert result["execution_result"]["message"] == "已为你生成结果"
    assert result["execution_result"]["tool_call_history"] == []
    assert "args_schema" in provider.calls[0]["messages"][0]["content"]
    planner_context = provider.calls[0]["messages"][1]["content"]
    assert "可用工具: - light_control: 控制灯光" in planner_context
    assert "args_schema: {}" in planner_context


def test_agent_execution_loops_reason_then_final_output(monkeypatch) -> None:
    provider = SequenceProvider(
        [
            '{"step_type":"reason","is_done":false,"reason_summary":"先思考一下"}',
            '{"step_type":"final_output","is_done":true,"final_output":{"message":"执行完成"}}',
        ]
    )
    monkeypatch.setattr(agent_execution_module, "_get_powerful_provider", lambda: provider)
    monkeypatch.setattr(agent_execution_module, "get_mcp_client", lambda: None)

    result = asyncio.run(
        agent_execution_function(
            {
                "user_input": "帮我开灯",
                "conversation_text": "user: 帮我开灯",
                "intent": {"name": "device_control"},
                "selected_tools": [{"name": "light_control", "description": "控制灯光"}],
            }
        )
    )

    assert result["status"] == "completed"
    assert result["execution_result"]["message"] == "执行完成"
    assert len(provider.calls) == 2
    second_context = provider.calls[1]["messages"][1]["content"]
    assert "执行历史: [{'step_index': 1, 'step_type': 'reason', 'reason_summary': '先思考一下'}]" in second_context
    assert "当前步数: 1 / 最多步数: 6" in second_context


def test_agent_execution_fails_when_step_limit_is_exceeded(monkeypatch) -> None:
    provider = SequenceProvider(['{"step_type":"reason","is_done":false,"reason_summary":"继续思考"}'])
    monkeypatch.setattr(agent_execution_module, "_get_powerful_provider", lambda: provider)
    monkeypatch.setattr(agent_execution_module, "get_mcp_client", lambda: None)
    monkeypatch.setattr(agent_execution_module, "DEFAULT_STEP_LIMIT", 1)

    result = asyncio.run(
        agent_execution_function(
            {
                "user_input": "帮我开灯",
                "conversation_text": "user: 帮我开灯",
                "intent": {"name": "device_control"},
                "selected_tools": [{"name": "light_control", "description": "控制灯光"}],
            }
        )
    )

    assert result["status"] == "failed"
    assert result["execution_result"]["unfinished"] is True
    assert result["execution_result"]["message"] == "任务暂时无法完成，请稍后重试。"
    assert result["error"]["type"] == "agent_step_limit_exceeded"


def test_agent_execution_records_toolcall_summary_without_raw_observation(monkeypatch) -> None:
    provider = SequenceProvider(
        [
            '{"step_type":"toolcall","is_done":false,"tool_name":"light_control","tool_args":{"entity_id":"light.bedroom"}}',
            '{"step_type":"final_output","is_done":true,"final_output":{"message":"灯已打开"}}',
        ]
    )
    tool = FakeInvokableTool(
        "light_control",
        "控制灯光",
        {"entity_id": {"type": "string"}},
        result={"ok": True, "entity_id": "light.bedroom"},
    )
    monkeypatch.setattr(agent_execution_module, "_get_powerful_provider", lambda: provider)
    monkeypatch.setattr(agent_execution_module, "get_mcp_client", lambda: FakeMCPClient([tool]))

    result = asyncio.run(
        agent_execution_function(
            {
                "user_input": "帮我打开卧室灯",
                "conversation_text": "user: 帮我打开卧室灯",
                "intent": {"name": "device_control"},
                "selected_tools": [{"name": "light_control", "description": "控制灯光"}],
            }
        )
    )

    history_entry = result["execution_result"]["tool_call_history"][0]
    assert result["status"] == "completed"
    assert tool.calls == [{"entity_id": "light.bedroom"}]
    assert history_entry["tool_name"] == "light_control"
    assert "result_summary" in history_entry
    assert "observation" not in history_entry
    assert '"entity_id"' in provider.calls[0]["messages"][1]["content"]


def test_agent_execution_replans_once_after_tool_failure(monkeypatch) -> None:
    provider = SequenceProvider(
        [
            '{"step_type":"toolcall","is_done":false,"tool_name":"light_control","tool_args":{"entity_id":"light.one"}}',
            '{"step_type":"toolcall","is_done":false,"tool_name":"light_control","tool_args":{"entity_id":"light.two"}}',
        ]
    )
    tool = FakeInvokableTool(
        "light_control",
        "控制灯光",
        {"entity_id": {"type": "string"}},
        error=RuntimeError("ha failed"),
    )
    monkeypatch.setattr(agent_execution_module, "_get_powerful_provider", lambda: provider)
    monkeypatch.setattr(agent_execution_module, "get_mcp_client", lambda: FakeMCPClient([tool]))

    result = asyncio.run(
        agent_execution_function(
            {
                "user_input": "帮我打开卧室灯",
                "conversation_text": "user: 帮我打开卧室灯",
                "intent": {"name": "device_control"},
                "selected_tools": [{"name": "light_control", "description": "控制灯光"}],
            }
        )
    )

    assert result["status"] == "failed"
    assert result["error"]["type"] == "tool_execution_error"
    assert result["execution_result"]["message"] == "任务暂时无法完成，请稍后重试。"
    assert "RuntimeError: ha failed" in result["error"]["message"]
    assert len(result["execution_result"]["tool_call_history"]) == 2
    assert tool.calls == [{"entity_id": "light.one"}, {"entity_id": "light.two"}]


def test_agent_execution_dry_run_skips_real_tool_invocation(monkeypatch, caplog) -> None:
    provider = SequenceProvider(
        [
            '{"step_type":"toolcall","is_done":false,"tool_name":"light_control","tool_args":{"entity_id":"light.bedroom"}}',
            '{"step_type":"final_output","is_done":true,"final_output":{"message":"dry run done"}}',
        ]
    )
    tool = FakeInvokableTool(
        "light_control",
        "控制灯光",
        {"entity_id": {"type": "string"}},
        result={"ok": True},
    )
    monkeypatch.setenv("RA_DRY_RUN", "1")
    monkeypatch.setattr(agent_execution_module, "_get_powerful_provider", lambda: provider)
    monkeypatch.setattr(agent_execution_module, "get_mcp_client", lambda: FakeMCPClient([tool]))

    with caplog.at_level("INFO"):
        result = asyncio.run(
            agent_execution_function(
                {
                    "user_input": "帮我打开卧室灯",
                    "conversation_text": "user: 帮我打开卧室灯",
                    "intent": {"name": "device_control"},
                    "selected_tools": [{"name": "light_control", "description": "控制灯光"}],
                }
            )
        )

    history_entry = result["execution_result"]["tool_call_history"][0]
    assert result["status"] == "completed"
    assert tool.calls == []
    assert history_entry["tool_name"] == "light_control"
    assert "RA_DRY_RUN enabled; skipped real MCP invocation." in history_entry["result_summary"]
    assert "RA_DRY_RUN intercepted MCP tool call" in caplog.text



#!/usr/bin/env python3
"""
ESP32 BLE Beacon配置代码生成器

根据Room Agent配置生成ESP32端的beacon配置代码
"""

import sys
from pathlib import Path

# 添加项目路径
project_root = Path(__file__).parent.parent
sys.path.insert(0, str(project_root))

import yaml
import datetime
from room_mapping import ROOM_NAMES, get_major_for_room


def generate_esp32_beacon_config(room_config_path: str) -> str:
    """生成ESP32 beacon配置代码（自定义厂商数据格式）

    Args:
        room_config_path: Room Agent配置文件路径

    Returns:
        str: 生成的C代码片段，可直接替换main.c中的raw_data数组
    """
    # 加载Room Agent配置
    config_path = Path(project_root) / room_config_path
    with open(config_path, 'r') as f:
        room_config = yaml.safe_load(f)

    # 提取beacon配置
    agent_config = room_config.get("agent", {})
    beacon_config = room_config.get("beacon", {})

    room_id = agent_config.get("room_id")
    room_name_cn = ROOM_NAMES.get(room_id, room_id)
    major = beacon_config.get("major")
    minor = beacon_config.get("minor", 0)  # 用作capability bitmap
    measured_power = beacon_config.get("measured_power", -59)

    # 生成设备名（用于nRF Connect可见性）
    # BLE设备名最长8字节，从room_id截取
    if len(room_id) > 8:
        device_name = room_id[:8].capitalize()
    else:
        device_name = room_id.capitalize()

    # 生成C代码片段（替换main.c中的adv_raw_data和scan_rsp_raw_data）
    code = f"""/*
 * ESP32 BLE Beacon Raw Data配置
 * 自动生成 - 基于: {room_config_path}
 * 房间: {room_name_cn} ({room_id})
 * Major: {major}
 * 生成时间: {datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}
 *
 * 使用方法：
 * 1. 将下面的数组定义复制到 esp32-ble-beacon/main/main.c
 * 2. 替换原有的 adv_raw_data[] 和 scan_rsp_raw_data[] 数组
 * 3. 重新编译并烧录ESP32
 */

// ========== Advertising Raw Data ==========
static uint8_t adv_raw_data[] = {{
    // Flags (3字节)
    0x02, ESP_BLE_AD_TYPE_FLAG, 0x06,

    // Device Name (用于nRF Connect可见性)
    // Length: 1(type) + {len(device_name)}(name) = {len(device_name) + 1}
    0x{len(device_name) + 1:02X}, ESP_BLE_AD_TYPE_NAME_CMPL,
    {', '.join(f"'{c}'" for c in device_name)},

    // Manufacturer Specific Data (11字节)
    // Length: 1(type) + 2(company_id) + 1(beacon_type) + 1(version) + 4(agent_id) + 1(cap) + 1(status) = 11
    0x0B, ESP_BLE_AD_MANUFACTURER_SPECIFIC_TYPE,
    0xFF, 0xFF,                        // Company ID (private)
    0x01,                              // Beacon type: 0x01 = Room Agent
    0x01,                              // Protocol version
    // Agent ID (Major值，小端序uint32)
    0x{major & 0xFF:02X}, 0x{(major >> 8) & 0xFF:02X}, 0x{(major >> 16) & 0xFF:02X}, 0x{(major >> 24) & 0xFF:02X},
    // Capability bitmap (从minor读取)
    0x{minor & 0xFF:02X},
    // Agent status (0x00 = 正常)
    0x00,

    // TX Power (3字节)
    0x02, ESP_BLE_AD_TYPE_TX_PWR, 0x{measured_power & 0xFF:02X},
}};

// ========== Scan Response Raw Data ==========
static uint8_t scan_rsp_raw_data[] = {{
    // Device Address (8字节) - 会在运行时自动填充
    0x08, ESP_BLE_AD_TYPE_LE_DEV_ADDR,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

    // URI (可选，用于nRF Connect测试)
    // URI_PREFIX_NONE = "" (空字符串)
    // 完整URI: //wskun.top
    // Length: 1(type) + 1(prefix) + len("//wskun.top") = 13
    0x0D, ESP_BLE_AD_TYPE_URI, URI_PREFIX_NONE,
    '/', '/', 'w', 's', 'k', 'u', 'n', '.', 't', 'o', 'p',
}};

/*
 * 配置说明：
 *
 * 1. adv_raw_data 包含：
 *    - Flags: 0x06 (General Discoverable + BR/EDR not supported)
 *    - Device Name: {device_name}
 *    - Manufacturer Data: 自定义厂商数据
 *      - Company ID: 0xFFFF (私有)
 *      - Beacon Type: 0x01 (Room Agent)
 *      - Protocol Version: 0x01
 *      - Agent ID: {major} (对应{room_name_cn})
 *      - Capability: 0x{minor:02X}
 *      - Status: 0x00 (正常)
 *    - TX Power: {measured_power} dBm
 *
 * 2. scan_rsp_raw_data 包含：
 *    - Device Address: 运行时自动填充ESP32的MAC地址
 *    - URI: /{room_id} (用于测试)
 *
 * 3. 广播间隔在 main.c 的 adv_params 中设置（默认20ms)
 */

"""

    return code


def generate_esp32_main_snippet(room_config_path: str) -> str:
    """生成ESP32主程序代码片段

    Args:
        room_config_path: Room Agent配置文件路径

    Returns:
        str: 生成的C代码片段
    """
    # 加载配置
    config_path = Path(project_root) / room_config_path
    with open(config_path, 'r') as f:
        room_config = yaml.safe_load(f)

    agent_config = room_config.get("agent", {})
    beacon_config = room_config.get("beacon", {})

    room_id = agent_config.get("room_id")
    major = beacon_config.get("major")

    code = f"""/*
 * ESP32主程序 - BLE Beacon广播
 * 房间: {room_id}
 */

#include "esp_log.h"
#include "esp32_beacon_config.h"  // 使用生成的配置头文件
#include "esp_gap_ble_api.h"
#include "esp_wifi.h"

static const char *TAG = "ESP32_BEACON";

// 外设句柄
static uint16_t ble_beacon_gap_handle = 0;

// Beacon参数（从配置文件加载）
static esp_ble_beacon_data_t beacon_data = {{
    .set_min_adv_interval_ms = {beacon_config.get('interval', 1) * 1000},
    .set_max_adv_interval_ms = {beacon_config.get('interval', 1) * 1000},
}};

// 广播数据（iBeacon格式）
static uint8_t beacon_payload[25] = {{
    // Flags
    0x02, 0x01,
    // UUID (16字节)
    0x12, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
    0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
    // Major (房间ID) - 小端序
    (uint8_t)({major} & 0xFF),
    (uint8_t)({major} >> 8),
    // Minor (区域ID) - 小端序
    (uint8_t)({beacon_config.get('minor', 0)} & 0xFF),
    (uint8_t)({beacon_config.get('minor', 0)} >> 8),
    // Measured Power
    {beacon_config.get('measured_power', -59)}
}};

void app_main(void)
{{
    ESP_LOGI(TAG, "初始化ESP32 BLE Beacon...");
    ESP_LOGI(TAG, "房间ID: %u, Major: %u", {major}, {major});

    // 打印配置
    print_beacon_config();

    // 初始化BLE
    ESP_ERROR_CHECK(esp_nvic_alloc_irq_handler(BLE_DYNAMIC_IRQ, ESP_IRQ_PRIORITY_DEFAULT, NULL, NULL, 0));

    // 初始化GAP
    esp_ble_beacon_config_t ble_beacon_cfg = {{
        .beacon_type = BEACON_TYPE_IBEACON,
    }};

    ESP_ERROR_CHECK(
        esp_ble_beacon_config(&ble_beacon_cfg) == ESP_OK,
        "配置BLE Beacon失败"
    );

    // 设置广播参数
    ESP_ERROR_CHECK(
        esp_ble_beacon_start(&ble_beacon_gap_handle, &beacon_data) == ESP_OK,
        "启动BLE Beacon失败"
    );

    // 设置广播数据
    struct esp_ble_beacon_data  beacon_data_struct = {{
        .flag = 0x4,
        .uuid_size = 16,
        .uuid = beacon_config.uuid,
        .major = beacon_config.major,
        .minor = beacon_config.minor,
        .power = beacon_config.measured_power,
    }};

    ESP_ERROR_CHECK(
        esp_ble_beacon_set_data(&ble_beacon_gap_handle, &beacon_data_struct) == ESP_OK,
        "设置BLE Beacon数据失败"
    );

    ESP_LOGI(TAG, "BLE Beacon启动成功");
    ESP_LOGI(TAG, "正在广播beacon信号...");
    ESP_LOGI(TAG, "房间: %s, Major: %u, Minor: %u",
              "{room_id}", {major}, {beacon_config.get('minor', 0)});

    // 保持运行
    while (1) {{
        vTaskDelay(pdMS_TO_TICKS({beacon_config.get('interval', 1) * 1000}));
    }}
}}
"""

    return code


def generate_esp32_sdk_config(room_config_path: str) -> str:
    """生成ESP-IDF SDK配置文件

    Args:
        room_config_path: Room Agent配置文件路径

    Returns:
        str: sdkconfig.defaults内容
    """
    # 加载配置
    config_path = Path(project_root) / room_config_path
    with open(config_path, 'r') as f:
        room_config = yaml.safe_load(f)

    agent_config = room_config.get("agent", {})
    beacon_config = room_config.get("beacon", {})

    room_id = agent_config.get("room_id")
    major = beacon_config.get("major")

    config = f"""# ESP-IDF SDK配置文件
# 对应Room Agent配置: {room_config_path}

# 房间配置
CONFIG_ROOM_ID="{room_id}"
CONFIG_ROOM_MAJOR={major}
CONFIG_ROOM_MINOR={beacon_config.get('minor', 0)}

# BLE Beacon配置
CONFIG_BEACON_ENABLED=y
CONFIG_BEACON_INTERVAL={beacon_config.get('interval', 1)}

# WiFi配置（根据实际网络修改）
CONFIG_ESP_WIFI_SSID="YourWiFiSSID"
CONFIG_ESP_WIFI_PASSWORD="YourWiFiPassword"

# MQTT配置（可选：ESP32作为MQTT客户端）
CONFIG_MQTT_ENABLED=n
# CONFIG_MQTT_BROKER_URI="mqtt://192.168.1.100:1883"

# 日志级别
CONFIG_LOG_DEFAULT_LEVEL_INFO=1
"""

    return config


def main():
    """主函数"""
    import argparse

    parser = argparse.ArgumentParser(
        description="生成ESP32 BLE Beacon配置代码"
    )
    parser.add_argument(
        "--config",
        default="config/room_agent.yaml",
        help="Room Agent配置文件路径"
    )
    parser.add_argument(
        "--output",
        help="输出目录（默认：esp32_beacon_config/）"
    )
    parser.add_argument(
        "--type",
        choices=["header", "main", "sdkconfig", "all"],
        default="all",
        help="生成的代码类型"
    )

    args = parser.parse_args()

    # 确定输出目录
    if args.output:
        output_dir = Path(args.output)
    else:
        output_dir = Path(project_root) / "esp32_beacon_config"

    output_dir.mkdir(parents=True, exist_ok=True)

    print(f"📄 读取配置: {args.config}")
    print(f"📝 输出目录: {output_dir}")

    # 生成代码
    if args.type in ["header", "all"]:
        header_code = generate_esp32_beacon_config(args.config)
        header_path = output_dir / "esp32_beacon_config.h"
        with open(header_path, 'w') as f:
            f.write(header_code)
        print(f"✅ 生成头文件: {header_path}")

    # 不再生成main_beacon.c，因为用户只需要复制raw_data数组到esp32-ble-beacon项目

    if args.type in ["sdkconfig", "all"]:
        sdk_config = generate_esp32_sdk_config(args.config)
        sdk_path = output_dir / "sdkconfig.defaults"
        with open(sdk_path, 'w') as f:
            f.write(sdk_config)
        print(f"✅ 生成SDK配置: {sdk_path}")

    # 加载配置用于README生成
    config_path = Path(project_root) / args.config
    with open(config_path, 'r') as f:
        room_config = yaml.safe_load(f)

    agent_config = room_config.get("agent", {})
    beacon_config = room_config.get("beacon", {})

    room_id = agent_config.get("room_id")
    room_name_cn = ROOM_NAMES.get(room_id, room_id)
    major = beacon_config.get("major")
    minor = beacon_config.get("minor", 0)
    measured_power = beacon_config.get("measured_power", -59)

    # 生成README
    readme_content = f"""# ESP32 BLE Beacon配置文件

## 生成时间
{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}

## 对应Room Agent配置
{args.config}

## 文件说明

### esp32_beacon_config.h
包含自定义厂商数据格式的`raw_data`数组定义：
- `adv_raw_data[]` - 广播数据数组
- `scan_rsp_raw_data[]` - 扫描响应数据数组

## 使用方法

### 步骤1: 复制数组到ESP32项目

打开 `esp32_beacon_config.h`，复制其中的两个数组：
- `adv_raw_data[]`
- `scan_rsp_raw_data[]`

粘贴到 `esp32-ble-beacon/main/main.c`，替换原有的同名数组。

### 步骤2: 编译ESP32项目

```bash
cd /path/to/esp32-ble-beacon

# 清理之前的编译
idf.py fullclean

# 配置项目
idf.py reconfigure

# 编译
idf.py build
```

### 步骤3: 烧录到ESP32

```bash
# 烧录（根据实际端口修改）
idf.py -p /dev/ttyUSB0 flash

# 监视串口（可选）
idf.py -p /dev/ttyUSB0 monitor
```

## 验证Beacon

使用BLE扫描工具验证：

### 使用nRF Connect（手机App）
1. 打开nRF Connect
2. 扫描设备
3. 找到名称为 "{room_id.capitalize()}" 的设备
4. 查看厂商数据：
   - Company ID: 0xFFFF
   - Beacon Type: 0x01
   - Agent ID: {major}

### 使用命令行（Linux）
```bash
hcitool lescan
```

### 使用命令行（macOS）
```bash
bleutil scan
```

## 配置详情

### Beacon数据格式
- **Company ID**: 0xFFFF (私有)
- **Beacon Type**: 0x01 (Room Agent)
- **Protocol Version**: 0x01
- **Agent ID**: {major} (对应{room_name_cn})
- **Capability**: 0x{minor:02X}
- **Status**: 0x00 (正常)
- **TX Power**: {measured_power} dBm

### 联动Room Agent

1. 确保Room Agent配置文件正确
2. 运行验证脚本：
   ```bash
   python3 scripts/validate_beacon_binding.py --config {args.config}
   ```
3. 启动Room Agent：
   ```bash
   python3 main_room_agent.py
   ```

## 故障排查

### 问题：扫描不到beacon
1. 检查ESP32供电
2. 检查ESP32固件是否正常运行
3. 使用串口监视查看ESP32日志

### 问题：Agent ID不匹配
1. 检查生成的`adv_raw_data`中的Agent ID
2. 检查Room Agent的`beacon.major`值
3. 运行验证脚本

### 问题：RSSI信号弱
1. 调整`measured_power`校准值
2. 减少beacon与接收器距离
3. 检查ESP32发射功率设置
"""

    readme_path = output_dir / "README.md"
    with open(readme_path, 'w', encoding='utf-8') as f:
        f.write(readme_content)
    print(f"✅ 生成README: {readme_path}")

    print(f"\n✅ 配置文件生成完成！")
    print(f"📁 输出目录: {output_dir.absolute()}")
    print(f"\n下一步：")
    print(f"  1. 查看README: {readme_path}")
    print(f"  2. 复制文件到ESP32项目")
    print(f"  3. 编译并烧录ESP32")
    print(f"  4. 验证beacon信号")
    print(f"  5. 运行验证脚本：python3 scripts/validate_beacon_binding.py")


if __name__ == "__main__":
    import datetime
    sys.exit(main())



"""Graph nodes for the Room Agent workflow."""

from .direct_response import direct_response
from .intent_recognition import intent_recognition
from .tool_selection import tool_selection

__all__ = ["direct_response", "intent_recognition", "tool_selection"]



"""Minimal CLI entrypoint for the LangGraph rebuild."""

from __future__ import annotations

import argparse
import asyncio
import json
from typing import Any

from app.server import initialize_runtime_dependencies
from config.settings import LLMRole, load_settings
from graph.entry import compile_graph
from integrations.llm_provider import create_llm_provider_registry


async def run_once(
    user_input: str,
    config_path: str | None = None,
    llm_config_path: str | None = None,
) -> dict[str, Any]:
    from integrations.mcp_client import build_home_assistant_mcp_client
    from app.server import probe_home_assistant_mcp

    settings = load_settings(config_path, llm_config_path)
    llm_registry = create_llm_provider_registry(settings.llm)
    if llm_registry.get(LLMRole.LOW_COST) is None:
        raise ValueError(
            "Low-cost LLM provider is unavailable. "
            "Check the llm config and ensure the low_cost role has valid credentials."
        )

    mcp_client = build_home_assistant_mcp_client(settings.agent.home_assistant_mcp)
    mcp_health_status = await probe_home_assistant_mcp(
        mcp_client,
        settings.agent.home_assistant_mcp,
    )

    initialize_runtime_dependencies(
        settings=settings,
        llm_provider_registry=llm_registry,
        mcp_client=mcp_client,
        mcp_health_status=mcp_health_status,
    )
    graph = compile_graph()

    final_state = await graph.ainvoke({"user_input": user_input})
    return final_state


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Run the rebuilt room-agent graph once.")
    parser.add_argument("user_input", help="User input to send into the graph.")
    parser.add_argument("--config", dest="config_path", required=True, help="YAML config path.")
    parser.add_argument(
        "--llm-config", dest="llm_config_path", required=True, help="LLM config path."
    )
    return parser


def main() -> None:
    args = _build_parser().parse_args()
    result = asyncio.run(
        run_once(
            user_input=args.user_input,
            config_path=args.config_path,
            llm_config_path=args.llm_config_path,
        )
    )
    print(json.dumps(result, ensure_ascii=False, indent=2))


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
from uuid import uuid4
from unittest.mock import Mock

from a2a.server.agent_execution import RequestContext
from a2a.server.events import EventQueue
from a2a.types import Message, MessageSendParams, Part, Role, Task, TaskState, TaskStatus, TextPart

from app.a2a_server import A2A_FALLBACK_RESPONSE, RoomAgentExecutor, _build_conversation_text


def _message(
    text: str,
    *,
    role: Role,
    task_id: str = "task-1",
    context_id: str = "ctx-1",
) -> Message:
    return Message(
        role=role,
        message_id=str(uuid4()),
        task_id=task_id,
        context_id=context_id,
        parts=[Part(root=TextPart(text=text))],
    )


def test_build_conversation_text_renders_history_and_current_input() -> None:
    history = [
        _message("你好", role=Role.user),
        _message("你好，我在。", role=Role.agent),
        _message("帮我打开卧室灯", role=Role.user),
    ]
    task = Task(
        id="task-1",
        context_id="ctx-1",
        status=TaskStatus(state=TaskState.submitted),
        history=history,
    )

    conversation_text = _build_conversation_text(
        task=task,
        current_message=history[-1],
    )

    assert "user: 你好" in conversation_text
    assert "assistant: 你好，我在。" in conversation_text
    assert "user: 帮我打开卧室灯" in conversation_text
    assert "Current user input:\n帮我打开卧室灯" in conversation_text


def test_invoke_roomagent_entrypoint_falls_back_when_graph_message_missing(
    monkeypatch,
) -> None:
    executor = RoomAgentExecutor()
    logger = Mock()

    class FakeGraph:
        async def ainvoke(self, payload):
            assert payload["conversation_text"] == "user: 你好"
            return {"execution_result": {"type": "placeholder"}}

    monkeypatch.setattr("app.a2a_server._compile_graph", lambda: FakeGraph())
    monkeypatch.setattr("app.a2a_server.logger", logger)

    result = asyncio.run(
        executor.invoke_roomagent_entrypoint(
            user_input="你好",
            context_id="ctx-1",
            task_id="task-1",
            conversation_text="user: 你好",
        )
    )

    assert result == A2A_FALLBACK_RESPONSE
    logger.info.assert_called_once()
    assert '"type": "placeholder"' in logger.info.call_args.args[3]


def test_invoke_roomagent_entrypoint_returns_sanitized_graph_message(monkeypatch) -> None:
    executor = RoomAgentExecutor()

    class FakeGraph:
        async def ainvoke(self, payload):
            assert payload["conversation_text"] == "user: 帮我开灯"
            return {
                "error": {
                    "type": "tool_execution_error",
                    "message": "PermissionError: 401 Unauthorized for home_assistant",
                },
                "execution_result": {
                    "type": "agent_execution_unfinished",
                    "message": "任务暂时无法完成，请稍后重试。",
                    "unfinished": True,
                },
            }

    monkeypatch.setattr("app.a2a_server._compile_graph", lambda: FakeGraph())

    result = asyncio.run(
        executor.invoke_roomagent_entrypoint(
            user_input="帮我开灯",
            context_id="ctx-1",
            task_id="task-1",
            conversation_text="user: 帮我开灯",
        )
    )

    assert result == "任务暂时无法完成，请稍后重试。"


def test_execute_enqueues_artifact_for_new_task(monkeypatch) -> None:
    executor = RoomAgentExecutor()
    request_message = _message("你好", role=Role.user)
    context = RequestContext(request=MessageSendParams(message=request_message))
    event_queue = EventQueue()

    async def fake_invoke_roomagent_entrypoint(**kwargs):
        assert kwargs["user_input"] == "你好"
        assert "Current user input:\n你好" in kwargs["conversation_text"]
        return "graph reply"

    monkeypatch.setattr(
        executor,
        "invoke_roomagent_entrypoint",
        fake_invoke_roomagent_entrypoint,
    )

    async def _run() -> list[object]:
        await executor.execute(context, event_queue)
        events: list[object] = []
        while True:
            try:
                events.append(await event_queue.dequeue_event(no_wait=True))
            except asyncio.QueueEmpty:
                return events

    events = asyncio.run(_run())

    assert len(events) == 3
    task_event, artifact_event, status_event = events
    assert isinstance(task_event, Task)
    assert artifact_event.artifact.parts[0].root.text == "graph reply"
    assert status_event.status.state == TaskState.completed



"""Async service entrypoint for the RoomAgent runtime."""

from __future__ import annotations

import asyncio
import contextlib
import logging
import os
import argparse
from dataclasses import asdict, dataclass
from datetime import datetime, timezone

import uvicorn

from .a2a_server import build_a2a_application
from config.settings import HomeAssistantMCPSettings, LLMRole, Settings, load_settings
from .gateway_client import GatewayClient
from integrations.mcp_client import MCPToolClient, build_home_assistant_mcp_client
from integrations.llm_provider import LLMProviderRegistry, create_llm_provider_registry


logger = logging.getLogger(__name__)
DIRECT_EXECUTION_ERROR = (
    "Do not run room-agent/app/server.py directly. "
    "Use the project script entry instead, for example: "
    "`cd room-agent && uv run serve --config-path config/examples/room_agent.example.yaml "
    "--llm-config-path tests/fixtures/llm.yaml`."
)
_SETTINGS: Settings | None = None
_LLM_PROVIDER_REGISTRY: LLMProviderRegistry | None = None
_MCP_CLIENT: MCPToolClient | None = None
_CONFIG_PATH: str | None = None
_LLM_CONFIG_PATH: str | None = None


@dataclass(slots=True)
class MCPHealthStatus:
    enabled: bool = False
    healthy: bool = False
    server_name: str | None = None
    checked_at: str | None = None
    prompt_count: int | None = None
    error_type: str | None = None
    error: str | None = None


_MCP_HEALTH_STATUS = MCPHealthStatus()


def initialize_runtime_dependencies(
    *,
    settings: Settings,
    llm_provider_registry: LLMProviderRegistry | None = None,
    mcp_client: MCPToolClient | None = None,
    mcp_health_status: MCPHealthStatus | None = None,
) -> None:
    """Initialize process-wide runtime singletons explicitly."""
    global _SETTINGS
    global _LLM_PROVIDER_REGISTRY
    global _MCP_CLIENT
    global _MCP_HEALTH_STATUS

    _SETTINGS = settings
    _LLM_PROVIDER_REGISTRY = llm_provider_registry or create_llm_provider_registry(settings.llm)
    _MCP_CLIENT = mcp_client
    _MCP_HEALTH_STATUS = mcp_health_status or MCPHealthStatus()


def get_settings() -> Settings:
    """Return the process-wide settings singleton."""
    global _SETTINGS
    if _SETTINGS is None:
        _SETTINGS = load_settings(config_path=_CONFIG_PATH, llm_config_path=_LLM_CONFIG_PATH)
    return _SETTINGS


def get_mcp_client() -> MCPToolClient | None:
    """Return the process-wide MCP client singleton."""
    return _MCP_CLIENT


def get_mcp_health_status() -> dict[str, object | None]:
    """Return the last known Home Assistant MCP health status."""
    return asdict(_MCP_HEALTH_STATUS)


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments for runtime configuration."""
    parser = argparse.ArgumentParser(description="Run RoomAgent runtime service.")
    parser.add_argument(
        "--config-path",
        dest="config_path",
        required=True,
        help="Path to room-agent main config file.",
    )
    parser.add_argument(
        "--llm-config-path",
        dest="llm_config_path",
        required=True,
        help="Path to room-agent LLM config file.",
    )
    return parser.parse_args()


def get_llm_provider_registry() -> LLMProviderRegistry:
    """Return the process-wide LLM provider registry singleton."""
    global _LLM_PROVIDER_REGISTRY
    if _LLM_PROVIDER_REGISTRY is None:
        settings = get_settings()
        _LLM_PROVIDER_REGISTRY = create_llm_provider_registry(settings.llm)
    return _LLM_PROVIDER_REGISTRY


def _describe_exception(exc: BaseException) -> tuple[str, str]:
    """Format exception details for structured health-check logging."""
    error_type = type(exc).__name__
    message = f"{error_type}: {exc}"

    if isinstance(exc, BaseExceptionGroup):
        nested_messages = [_describe_exception(item)[1] for item in exc.exceptions]
        if nested_messages:
            message = f"{message} | nested=[{'; '.join(nested_messages)}]"

    return error_type, message


async def probe_home_assistant_mcp(
    client: MCPToolClient | None,
    settings: HomeAssistantMCPSettings | None,
) -> MCPHealthStatus:
    """Probe prompt capability without interrupting room-agent startup."""
    status = MCPHealthStatus(
        enabled=bool(settings and settings.enabled),
        server_name=settings.server_name if settings else None,
        checked_at=datetime.now(timezone.utc).isoformat(),
    )
    if settings is None or not settings.enabled:
        return status
    if client is None:
        status.error = "Home Assistant MCP is enabled but client was not initialized."
        return status
    if not settings.health_check.enabled:
        status.healthy = True
        return status

    try:
        result = await client.list_prompts(settings.server_name)
        prompts = getattr(result, "prompts", None)
        status.prompt_count = len(prompts) if prompts is not None else 0
        status.healthy = True
    except Exception as exc:
        status.error_type, status.error = _describe_exception(exc)
    return status


@dataclass(slots=True)
class ServiceRuntime:
    """Coordinates the long-lived tasks hosted by the RoomAgent process."""

    host: str = os.getenv("ROOM_AGENT_HOST", "127.0.0.1")
    port: int = int(os.getenv("ROOM_AGENT_PORT", "10000"))
    business_poll_interval_seconds: float = 30.0
    gateway_client: GatewayClient | None = None

    async def run(self) -> None:
        """Run the HTTP server loop and the business loop in one event loop."""
        global _MCP_CLIENT
        global _MCP_HEALTH_STATUS

        settings = get_settings()
        llm_registry = get_llm_provider_registry()
        _MCP_CLIENT = build_home_assistant_mcp_client(settings.agent.home_assistant_mcp)
        _MCP_HEALTH_STATUS = await probe_home_assistant_mcp(
            _MCP_CLIENT,
            settings.agent.home_assistant_mcp,
        )
        gateway_client = self.gateway_client
        if settings.agent.gateway and gateway_client is None:
            gateway_client = GatewayClient()
            self.gateway_client = gateway_client
        logger.info(
            "RoomAgent settings loaded for agent=%s room=%s powerful_provider=%s low_cost_provider=%s",
            settings.agent.id,
            settings.agent.room_id,
            (
                type(llm_registry.get(LLMRole.POWERFUL)).__name__
                if llm_registry.get(LLMRole.POWERFUL)
                else "None"
            ),
            (
                type(llm_registry.get(LLMRole.LOW_COST)).__name__
                if llm_registry.get(LLMRole.LOW_COST)
                else "None"
            ),
        )
        logger.info(
            "Home Assistant MCP status enabled=%s healthy=%s server=%s prompt_count=%s error_type=%s error=%s",
            _MCP_HEALTH_STATUS.enabled,
            _MCP_HEALTH_STATUS.healthy,
            _MCP_HEALTH_STATUS.server_name,
            _MCP_HEALTH_STATUS.prompt_count,
            _MCP_HEALTH_STATUS.error_type,
            _MCP_HEALTH_STATUS.error,
        )

        if (
            settings.agent.gateway
            and settings.agent.gateway.register_on_startup
            and gateway_client is not None
        ):
            await gateway_client.register_agent(settings.agent)

        stop_event = asyncio.Event()
        tasks = [
            asyncio.create_task(self.run_a2a_http_server(stop_event), name="roomagent-a2a-http"),
            asyncio.create_task(self.run_business_loop(stop_event), name="roomagent-business-loop"),
        ]
        if settings.agent.gateway and gateway_client is not None:
            tasks.append(
                asyncio.create_task(
                    self.run_gateway_heartbeat_loop(stop_event),
                    name="roomagent-gateway-heartbeat",
                )
            )

        try:
            await asyncio.gather(*tasks)
        except asyncio.CancelledError:
            logger.info("RoomAgent runtime cancelled, stopping background tasks.")
            await self._shutdown_tasks(tasks, stop_event)
            raise
        finally:
            if any(not task.done() for task in tasks):
                await self._shutdown_tasks(tasks, stop_event)

    async def _shutdown_tasks(
        self,
        tasks: list[asyncio.Task[None]],
        stop_event: asyncio.Event,
    ) -> None:
        """Stop background tasks, preferring graceful shutdown before cancellation."""
        stop_event.set()
        _, pending = await asyncio.wait(tasks, timeout=5)
        for task in pending:
            task.cancel()
        if pending:
            await asyncio.gather(*pending, return_exceptions=True)

    async def run_a2a_http_server(self, stop_event: asyncio.Event) -> None:
        """Run the RoomAgent A2A HTTP service until shutdown is requested."""
        server = uvicorn.Server(
            uvicorn.Config(
                build_a2a_application(host=self.host, port=self.port).build(),
                host=self.host,
                port=self.port,
                log_level="info",
            )
        )
        server_task = asyncio.create_task(server.serve(), name="roomagent-uvicorn")

        logger.info("RoomAgent A2A HTTP server started on http://%s:%s", self.host, self.port)
        try:
            await stop_event.wait()
        finally:
            server.should_exit = True
            with contextlib.suppress(asyncio.CancelledError):
                await server_task
            logger.info("RoomAgent A2A HTTP server stopped.")

    async def run_business_loop(self, stop_event: asyncio.Event) -> None:
        """Periodic business loop placeholder."""
        logger.info("Business loop placeholder started.")
        while not stop_event.is_set():
            await self.tick_business_jobs()
            try:
                await asyncio.wait_for(
                    stop_event.wait(),
                    timeout=self.business_poll_interval_seconds,
                )
            except TimeoutError:
                continue
        logger.info("Business loop placeholder stopped.")

    async def run_gateway_heartbeat_loop(self, stop_event: asyncio.Event) -> None:
        """Send heartbeat requests to the backend gateway on the configured interval."""
        settings = get_settings()
        gateway = settings.agent.gateway
        if gateway is None:
            return

        logger.info(
            "Gateway heartbeat loop started for agent=%s interval=%ss",
            settings.agent.id,
            gateway.heartbeat_interval,
        )
        while not stop_event.is_set():
            if self.gateway_client is not None:
                await self.gateway_client.send_heartbeat(settings.agent)
            try:
                await asyncio.wait_for(
                    stop_event.wait(),
                    timeout=gateway.heartbeat_interval,
                )
            except TimeoutError:
                continue
        logger.info("Gateway heartbeat loop stopped.")

    async def tick_business_jobs(self) -> None:
        """Reserved hook for future scheduled business queries."""
        return None


async def main() -> None:
    """Module-level async entrypoint."""
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s %(levelname)s [%(name)s] %(message)s",
    )
    await ServiceRuntime().run()


def cli() -> None:
    """Synchronous console-script entrypoint for packaging tools."""
    global _CONFIG_PATH
    global _LLM_CONFIG_PATH

    args = parse_args()
    _CONFIG_PATH = args.config_path
    _LLM_CONFIG_PATH = args.llm_config_path
    with contextlib.suppress(KeyboardInterrupt):
        asyncio.run(main())


if __name__ == "__main__":
    raise SystemExit(DIRECT_EXECUTION_ERROR)



"""Agent-execution subgraph for multi-step tool-enabled execution."""

from __future__ import annotations

import json
import logging
import os
from typing import Any

from langchain_core.tools import BaseTool
from langgraph.graph import END, START, StateGraph

from app.a2a_server import create_text_part, get_current_updater
from a2a.types import TaskState
from app.server import get_llm_provider_registry, get_mcp_client
from config.settings import LLMRole
from graph.state import RoomAgentGraphState
from llm_json_parse import JsonParserWithRepair

from .state import AgentExecutionState, FinalOutput, PlannerStep


logger = logging.getLogger(__name__)
DEFAULT_STEP_LIMIT = 6
USER_VISIBLE_UNFINISHED_MESSAGE = "任务暂时无法完成, 请稍后重试。"
PLANNER_OUTPUT_SCHEMA = {
    "type": "object",
    "properties": {
        "step_type": {
            "type": "string",
            "enum": ["reason", "toolcall", "final_output"],
        },
        "is_done": {"type": "boolean"},
        "reason_summary": {"type": "string"},
        "tool_name": {"type": "string"},
        "tool_args": {"type": "object"},
        "final_output": {
            "type": "object",
            "properties": {
                "message": {"type": "string", "minLength": 1},
                "summary": {"type": ["string", "null"]},
                "metadata": {"type": ["object", "null"]},
            },
            "required": ["message"],
            "additionalProperties": False,
        },
    },
    "required": ["step_type", "is_done"],
    "additionalProperties": False,
}


async def agent_execution(state: RoomAgentGraphState) -> RoomAgentGraphState:
    """Run the internal agent-execution subgraph and return an outer-state patch."""
    result = await compile_agent_execution_subgraph().ainvoke(
        {
            "user_input": state.get("user_input", ""),
            "conversation_text": state.get("conversation_text", ""),
            "intent": dict(state.get("intent", {})),
            "selected_tools": list(state.get("selected_tools", [])),
            "metadata": dict(state.get("metadata", {})),
        }
    )
    return result.get("outer_state_patch", {})


def compile_agent_execution_subgraph() -> Any:
    """Compile the internal agent-execution subgraph."""
    graph = StateGraph(AgentExecutionState)
    graph.add_node("subgraph_input_transform", subgraph_input_transform)
    graph.add_node("agent_plan_step", agent_plan_step)
    graph.add_node("agent_dispatch_step", agent_dispatch_step)
    graph.add_node("agent_record_reason", agent_record_reason)
    graph.add_node("agent_execute_toolcall", agent_execute_toolcall)
    graph.add_node("agent_finalize_output", agent_finalize_output)
    graph.add_node("agent_handle_limit_or_error", agent_handle_limit_or_error)
    graph.add_node("agent_submit_reply", agent_submit_reply)
    graph.add_node("subgraph_output_transform", subgraph_output_transform)

    graph.add_edge(START, "subgraph_input_transform")
    graph.add_edge("subgraph_input_transform", "agent_plan_step")
    graph.add_conditional_edges(
        "agent_plan_step",
        route_after_plan_step,
        {
            "dispatch": "agent_dispatch_step",
            "handle_error": "agent_handle_limit_or_error",
        },
    )
    graph.add_conditional_edges(
        "agent_dispatch_step",
        route_after_dispatch_step,
        {
            "reason": "agent_record_reason",
            "toolcall": "agent_execute_toolcall",
            "final_output": "agent_finalize_output",
            "handle_error": "agent_handle_limit_or_error",
        },
    )
    graph.add_edge("agent_record_reason", "agent_plan_step")
    graph.add_conditional_edges(
        "agent_execute_toolcall",
        route_after_toolcall,
        {
            "plan": "agent_plan_step",
            "handle_error": "agent_handle_limit_or_error",
        },
    )
    graph.add_edge("agent_finalize_output", "agent_submit_reply")
    graph.add_edge("agent_handle_limit_or_error", "agent_submit_reply")
    graph.add_edge("agent_submit_reply", "subgraph_output_transform")
    graph.add_edge("subgraph_output_transform", END)
    return graph.compile()


async def subgraph_input_transform(state: AgentExecutionState) -> AgentExecutionState:
    """Initialize internal state for the agent loop."""
    available_tools, tool_instances = await _load_selected_tools(state.get("selected_tools", []))
    return {
        "user_input": state.get("user_input", "").strip(),
        "conversation_text": state.get("conversation_text", "").strip()
        or state.get("user_input", "").strip(),
        "intent": dict(state.get("intent", {})),
        "metadata": dict(state.get("metadata", {})),
        "available_tools": available_tools,
        "available_tool_instances": tool_instances,
        "step_count": 0,
        "step_limit": DEFAULT_STEP_LIMIT,
        "step_history": [],
        "raw_model_outputs": [],
        "tool_results": [],
        "replan_used": False,
    }


async def agent_plan_step(state: AgentExecutionState) -> AgentExecutionState:
    """Plan the next agent step with the powerful model."""
    if state.get("step_count", 0) >= state.get("step_limit", DEFAULT_STEP_LIMIT):
        return {
            "loop_decision": "handle_error",
            "terminal_error": {
                "type": "agent_step_limit_exceeded",
                "message": (
                    f"Agent step limit exceeded before completion "
                    f"(limit={state.get('step_limit', DEFAULT_STEP_LIMIT)})."
                ),
                "source_node": "agent_plan_step",
                "retryable": False,
            },
        }

    provider = _get_powerful_provider()
    raw_output = await provider.complete_text(
        _build_planner_messages(state),
        temperature=0,
        json_mode=True,
    )
    parsed = await JsonParserWithRepair(llm_provider=provider)(
        raw_output,
        schema=PLANNER_OUTPUT_SCHEMA,
    )
    current_step = _normalize_planner_step(parsed)
    _validate_planner_step(current_step)

    raw_model_outputs = list(state.get("raw_model_outputs", []))
    raw_model_outputs.append(raw_output)

    return {
        "current_step": current_step,
        "raw_model_outputs": raw_model_outputs,
        "step_count": state.get("step_count", 0) + 1,
        "loop_decision": "dispatch",
    }


def agent_dispatch_step(state: AgentExecutionState) -> AgentExecutionState:
    """No-op node used to keep dispatch logic explicit in the graph."""
    current_step = state.get("current_step", {})
    if not current_step:
        return {
            "loop_decision": "handle_error",
            "terminal_error": {
                "type": "agent_dispatch_error",
                "message": "Planner did not produce a current_step.",
                "source_node": "agent_dispatch_step",
                "retryable": False,
            },
        }
    return {}


def agent_record_reason(state: AgentExecutionState) -> AgentExecutionState:
    """Record a reasoning-only step and continue the loop."""
    current_step = state.get("current_step", {})
    step_history = list(state.get("step_history", []))

    updater = None
    try:
        updater = get_current_updater()
    except RuntimeError:
        logger.debug("No TaskUpdater available; skip reply submit from agent_execution subgraph.")

    if updater and current_step.get("reason_summary"):
        message = updater.new_agent_message(
            parts=[create_text_part(current_step.get("reason_summary", "思考中...").strip())],
        )
        _ = updater.update_status(TaskState.working, message)

    step_history.append(
        {
            "step_index": state.get("step_count", 0),
            "step_type": "reason",
            "reason_summary": current_step.get("reason_summary", "").strip(),
        }
    )
    return {"step_history": step_history}


async def agent_execute_toolcall(state: AgentExecutionState) -> AgentExecutionState:
    """Execute a single tool call and either continue or fail."""
    current_step = state.get("current_step", {})
    tool_name = str(current_step.get("tool_name", "")).strip()
    tool_args = current_step.get("tool_args", {}) or {}
    if _is_dry_run_enabled():
        observation = _build_dry_run_observation(tool_name=tool_name, tool_args=tool_args)
        logger.info(
            "RA_DRY_RUN intercepted MCP tool call tool=%s observation=%s",
            tool_name,
            json.dumps(observation, ensure_ascii=False, default=str),
        )
        return _build_tool_success_state(
            state,
            tool_name=tool_name,
            tool_args=tool_args,
            observation=observation,
        )

    updater = None
    try:
        updater = get_current_updater()
    except RuntimeError:
        logger.debug("No TaskUpdater available; skip reply submit from agent_execution subgraph.")

    tool_instances = state.get("available_tool_instances", {})
    tool = tool_instances.get(tool_name)
    if tool is None:
        return _build_tool_failure_state(
            state,
            tool_name=tool_name,
            tool_args=tool_args,
            error_message=f"Tool '{tool_name}' is not available.",
            source_node="agent_execute_toolcall",
        )

    try:
        observation = await _invoke_tool(tool, tool_args)
    except Exception as exc:
        return _build_tool_failure_state(
            state,
            tool_name=tool_name,
            tool_args=tool_args,
            error_message=f"{type(exc).__name__}: {exc}",
            source_node="agent_execute_toolcall",
        )

    if updater:
        message = updater.new_agent_message(parts=[create_text_part(f"已执行 {tool_name}")])
        await updater.update_status(TaskState.working, message)

    return _build_tool_success_state(
        state,
        tool_name=tool_name,
        tool_args=tool_args,
        observation=observation,
    )


def agent_finalize_output(state: AgentExecutionState) -> AgentExecutionState:
    """Persist the final output and prepare to exit the subgraph."""
    current_step = state.get("current_step", {})
    final_output = current_step.get("final_output", {})
    step_history = list(state.get("step_history", []))
    step_history.append(
        {
            "step_index": state.get("step_count", 0),
            "step_type": "final_output",
            "message": final_output.get("message", ""),
            "summary": final_output.get("summary"),
        }
    )
    return {
        "final_output": final_output,
        "step_history": step_history,
    }


def agent_handle_limit_or_error(state: AgentExecutionState) -> AgentExecutionState:
    """Build terminal error state for an unfinished agent run."""
    terminal_error = dict(state.get("terminal_error", {}))
    if not terminal_error:
        terminal_error = {
            "type": "agent_execution_error",
            "message": "Agent execution failed without a structured error.",
            "source_node": "agent_handle_limit_or_error",
            "retryable": False,
        }
    return {"terminal_error": terminal_error}


async def agent_submit_reply(state: AgentExecutionState) -> AgentExecutionState:
    """Submit a user-visible reply through the active A2A updater when available."""
    reply = _resolve_terminal_reply_text(state)
    if not reply:
        return {}

    try:
        updater = get_current_updater()
    except RuntimeError:
        logger.debug("No TaskUpdater available; skip reply submit from agent_execution subgraph.")
        return {}

    message = updater.new_agent_message(parts=[create_text_part(reply)])
    await updater.complete(message)
    return {}


def subgraph_output_transform(state: AgentExecutionState) -> AgentExecutionState:
    """Map the internal terminal state back to the outer graph state."""
    tool_call_history = _build_tool_call_history(state.get("tool_results", []))
    final_output = _normalize_final_output(state.get("final_output"))
    terminal_error = state.get("terminal_error")

    if final_output and not terminal_error:
        execution_result = {
            "type": "agent_final_output",
            "message": final_output["message"],
            "tool_call_history": tool_call_history,
        }
        if final_output.get("summary") is not None:
            execution_result["summary"] = final_output.get("summary")
        if final_output.get("metadata") is not None:
            execution_result["metadata"] = final_output.get("metadata")
        outer_patch: RoomAgentGraphState = {
            "status": "completed",
            "next_action": "agent_execution",
            "execution_result": execution_result,
            "tool_call_history": tool_call_history,
        }
        return {"outer_state_patch": outer_patch}

    error_message = (
        str(terminal_error.get("message", "")).strip()
        if isinstance(terminal_error, dict)
        else "Agent execution did not complete."
    )
    outer_patch = {
        "status": "failed",
        "next_action": "agent_execution",
        "error": terminal_error
        or {
            "type": "agent_execution_error",
            "message": error_message or "Agent execution did not complete.",
            "source_node": "subgraph_output_transform",
            "retryable": False,
        },
        "execution_result": {
            "type": "agent_execution_unfinished",
            "message": USER_VISIBLE_UNFINISHED_MESSAGE,
            "tool_call_history": tool_call_history,
            "unfinished": True,
        },
        "tool_call_history": tool_call_history,
    }
    return {"outer_state_patch": outer_patch}


def _resolve_terminal_reply_text(state: AgentExecutionState) -> str:
    final_output = _normalize_final_output(state.get("final_output"))
    message = str(final_output.get("message", "")).strip()
    if message and not state.get("terminal_error"):
        return message
    return USER_VISIBLE_UNFINISHED_MESSAGE


def route_after_plan_step(state: AgentExecutionState) -> str:
    """Route after planning."""
    return state.get("loop_decision", "handle_error")


def route_after_dispatch_step(state: AgentExecutionState) -> str:
    """Dispatch the current planner step."""
    if state.get("loop_decision") == "handle_error":
        return "handle_error"
    current_step = state.get("current_step", {})
    return str(current_step.get("step_type", "handle_error"))


def route_after_toolcall(state: AgentExecutionState) -> str:
    """Route after a tool call finishes."""
    return state.get("loop_decision", "handle_error")


def _normalize_planner_step(parsed: dict[str, Any]) -> PlannerStep:
    step: PlannerStep = {
        "step_type": str(parsed["step_type"]),
        "is_done": bool(parsed["is_done"]),
    }
    if "reason_summary" in parsed:
        step["reason_summary"] = str(parsed.get("reason_summary") or "").strip()
    if "tool_name" in parsed:
        step["tool_name"] = str(parsed.get("tool_name") or "").strip()
    if "tool_args" in parsed and isinstance(parsed["tool_args"], dict):
        step["tool_args"] = parsed["tool_args"]
    if "final_output" in parsed and isinstance(parsed["final_output"], dict):
        step["final_output"] = _normalize_final_output(parsed["final_output"])
    return step


def _normalize_final_output(payload: dict[str, Any] | FinalOutput | None) -> FinalOutput:
    if not payload:
        return {}
    normalized: FinalOutput = {"message": str(payload.get("message", "")).strip()}
    if payload.get("summary") is not None:
        normalized["summary"] = str(payload.get("summary"))
    if payload.get("metadata") is not None and isinstance(payload.get("metadata"), dict):
        normalized["metadata"] = payload.get("metadata")
    return normalized


def _validate_planner_step(step: PlannerStep) -> None:
    step_type = step.get("step_type")
    is_done = bool(step.get("is_done"))
    if step_type == "reason":
        if is_done:
            raise ValueError("Planner step_type=reason must have is_done=false.")
        if not step.get("reason_summary"):
            raise ValueError("Planner step_type=reason requires reason_summary.")
        return
    if step_type == "toolcall":
        if is_done:
            raise ValueError("Planner step_type=toolcall must have is_done=false.")
        if not step.get("tool_name"):
            raise ValueError("Planner step_type=toolcall requires tool_name.")
        if not isinstance(step.get("tool_args", {}), dict):
            raise ValueError("Planner step_type=toolcall requires tool_args object.")
        return
    if step_type == "final_output":
        if not is_done:
            raise ValueError("Planner step_type=final_output must have is_done=true.")
        final_output = step.get("final_output", {})
        if not final_output.get("message"):
            raise ValueError("Planner step_type=final_output requires final_output.message.")
        return
    raise ValueError(f"Unsupported planner step_type={step_type!r}.")


def _render_tool_list(tools: list[dict[str, Any]]) -> str:
    if not tools:
        return "(none)"

    rendered_tools = []
    for tool in tools:
        rendered_tools.append(
            "\n".join(
                [
                    f"- {tool.get('name', '<unknown>')}: {tool.get('description') or 'No description'}",
                    f"  args_schema: {json.dumps(tool.get('args_schema', {}), ensure_ascii=False, sort_keys=True)}",
                ]
            )
        )
    return "\n".join(rendered_tools)


def _build_planner_messages(state: AgentExecutionState) -> list[dict[str, str]]:
    return [
        {
            "role": "system",
            "content": (
                "你是一个房间的智能管理助手。你的工作是管理房间的各种设备和事务, 根据用户的需求进行推理和决策。\n"
                "\n"
                "**你每轮的任务很简单：决定下一步该做什么。** 你有三个选择：\n"
                "\n"
                "1. **思考与分析**：如果问题比较复杂, 你可以先思考一下思路, 理清思路后再执行动作。简洁地说出你的想法即可。\n"
                "\n"
                "2. **执行工具**：当你知道需要做什么时, 就拿起一个可用的工具去执行。记住每次只能用一个工具。告诉系统工具的名称和所需参数。\n"
                "   你必须严格参考每个工具给出的 args_schema 来构造 tool_args。"
                "不要编造 schema 里没有的字段, 字段名和结构必须与 schema 对齐。\n"
                "\n"
                "3. **给出最终答案**：当任务完成或问题解决了, 就告诉用户结果。这是唯一一个真正结束对话的方式。\n"
                "\n"
                "**重要规则**: \n"
                "• 你最多可以进行 6 个步骤。如果无法完成, 就坦诚告诉用户。\n"
                "• 思考和工具调用都会让步骤计数器增加, 所以要高效地思考。\n"
                "• 如果工具执行失败, 你可以重新规划一次。再次失败的话就放弃吧。\n"
                "• 你必须只输出一个 JSON 对象, 不要输出额外解释、Markdown 或自然语言前后缀。\n"
                "• JSON 字段必须与系统要求一致：step_type、is_done, 以及按需提供 reason_summary、tool_name、tool_args、final_output。\n"
                "\n"
                "输出要求示例：\n"
                '{"step_type":"reason","is_done":false,"reason_summary":"先确认当前状态"}\n'
                '{"step_type":"toolcall","is_done":false,"tool_name":"climate_get_state","tool_args":{"entity_id":"climate.lab"}}\n'
                '{"step_type":"final_output","is_done":true,"final_output":{"message":"实验室空调当前正在运行"}}'
            ),
        },
        {
            "role": "user",
            "content": (
                f"用户的原始输入: {state.get('user_input', '')}\n"
                f"对话上下文: {state.get('conversation_text', '')}\n"
                f"识别的意图: {state.get('intent', {})}\n"
                f"可用工具: {_render_tool_list(state.get('available_tools', []))}\n"
                f"当前步数: {state.get('step_count', 0)} / 最多步数: {state.get('step_limit', DEFAULT_STEP_LIMIT)}\n"
                f"执行历史: {state.get('step_history', [])}\n"
                f"工具结果: {state.get('tool_results', [])}\n"
                f"已重新规划: {state.get('replan_used', False)}\n"
                "请只返回一个 JSON 对象。"
            ),
        },
    ]


def _get_powerful_provider() -> Any:
    provider = get_llm_provider_registry().get(LLMRole.POWERFUL)
    if provider is None:
        raise RuntimeError(f"LLM provider is unavailable for role={LLMRole.POWERFUL.value}")
    return provider


async def _load_selected_tools(
    selected_tools: list[dict[str, Any]],
) -> tuple[list[dict[str, Any]], dict[str, BaseTool]]:
    client = get_mcp_client()
    selected_names = {
        str(tool.get("name", "")).strip()
        for tool in selected_tools
        if isinstance(tool, dict) and str(tool.get("name", "")).strip()
    }
    if client is None:
        return list(selected_tools), {}

    tools = await client.get_tools()
    available_tools = []
    tool_instances: dict[str, BaseTool] = {}
    for tool in tools:
        if selected_names and tool.name not in selected_names:
            continue
        tool_instances[tool.name] = tool
        available_tools.append(
            {
                "name": tool.name,
                "description": tool.description or "",
                "args_schema": _extract_tool_schema(tool),
            }
        )
    if available_tools:
        return available_tools, tool_instances
    return list(selected_tools), {}


async def _invoke_tool(tool: BaseTool, tool_args: dict[str, Any]) -> Any:
    if hasattr(tool, "ainvoke"):
        return await tool.ainvoke(tool_args)
    if hasattr(tool, "arun"):
        return await tool.arun(tool_args)
    if hasattr(tool, "invoke"):
        return tool.invoke(tool_args)
    if hasattr(tool, "run"):
        return tool.run(tool_args)
    raise RuntimeError(f"Tool {getattr(tool, 'name', '<unknown>')} is not invokable.")


def _extract_tool_schema(tool: BaseTool) -> dict[str, Any]:
    get_input_schema = getattr(tool, "get_input_schema", None)
    if callable(get_input_schema):
        schema_model = get_input_schema()
        model_json_schema = getattr(schema_model, "model_json_schema", None)
        if callable(model_json_schema):
            schema = model_json_schema()
            return schema if isinstance(schema, dict) else {}

    args_schema = getattr(tool, "args", {}) or {}
    if args_schema:
        return {
            "type": "object",
            "properties": args_schema,
        }

    return {}


def _build_tool_failure_state(
    state: AgentExecutionState,
    *,
    tool_name: str,
    tool_args: dict[str, Any],
    error_message: str,
    source_node: str,
) -> AgentExecutionState:
    tool_results = list(state.get("tool_results", []))
    step_history = list(state.get("step_history", []))
    entry = {
        "step_index": state.get("step_count", 0),
        "tool_name": tool_name,
        "tool_args": tool_args,
        "error": error_message,
        "args_summary": _summarize_value(tool_args),
        "error_summary": _summarize_value(error_message),
    }
    tool_results.append(entry)
    step_history.append(
        {
            "step_index": state.get("step_count", 0),
            "step_type": "toolcall",
            "tool_name": tool_name,
            "args_summary": entry["args_summary"],
            "error_summary": entry["error_summary"],
        }
    )
    if not state.get("replan_used", False):
        return {
            "tool_results": tool_results,
            "step_history": step_history,
            "replan_used": True,
            "loop_decision": "plan",
        }

    return {
        "tool_results": tool_results,
        "step_history": step_history,
        "loop_decision": "handle_error",
        "terminal_error": {
            "type": "tool_execution_error",
            "message": error_message,
            "source_node": source_node,
            "retryable": False,
        },
    }


def _build_tool_success_state(
    state: AgentExecutionState,
    *,
    tool_name: str,
    tool_args: dict[str, Any],
    observation: Any,
) -> AgentExecutionState:
    tool_results = list(state.get("tool_results", []))
    step_history = list(state.get("step_history", []))
    result_summary = _summarize_value(observation)
    tool_results.append(
        {
            "step_index": state.get("step_count", 0),
            "tool_name": tool_name,
            "tool_args": tool_args,
            "observation": observation,
            "args_summary": _summarize_value(tool_args),
            "result_summary": result_summary,
        }
    )
    step_history.append(
        {
            "step_index": state.get("step_count", 0),
            "step_type": "toolcall",
            "tool_name": tool_name,
            "args_summary": _summarize_value(tool_args),
            "result_summary": result_summary,
        }
    )
    return {
        "tool_results": tool_results,
        "step_history": step_history,
        "loop_decision": "plan",
    }


def _build_tool_call_history(tool_results: list[dict[str, Any]]) -> list[dict[str, Any]]:
    history = []
    for entry in tool_results:
        item = {
            "step_index": entry.get("step_index"),
            "tool_name": entry.get("tool_name"),
            "args_summary": entry.get("args_summary", ""),
        }
        if entry.get("result_summary"):
            item["result_summary"] = entry["result_summary"]
        if entry.get("error_summary"):
            item["error_summary"] = entry["error_summary"]
        history.append(item)
    return history


def _summarize_value(value: Any, *, limit: int = 180) -> str:
    if isinstance(value, str):
        text = value
    else:
        try:
            text = json.dumps(value, ensure_ascii=False, default=str)
        except TypeError:
            text = str(value)
    text = text.strip()
    if len(text) <= limit:
        return text
    return text[: limit - 3] + "..."


def _is_dry_run_enabled() -> bool:
    value = os.getenv("RA_DRY_RUN", "").strip().lower()
    return value not in {"", "0", "false", "no", "off"}


def _build_dry_run_observation(*, tool_name: str, tool_args: dict[str, Any]) -> dict[str, Any]:
    return {
        "dry_run": True,
        "tool_name": tool_name,
        "tool_args": tool_args,
        "message": "RA_DRY_RUN enabled; skipped real MCP invocation.",
    }



"""A2A HTTP service building blocks for the RoomAgent runtime."""

from __future__ import annotations

import json
import logging
from typing import Any

from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.apps import A2AStarletteApplication
from a2a.server.events import EventQueue
from a2a.server.request_handlers import DefaultRequestHandler
from a2a.server.tasks import InMemoryTaskStore, TaskUpdater
from a2a.types import (
    AgentCapabilities,
    AgentCard,
    AgentSkill,
    Message,
    Part,
    Task,
    TaskState,
    TextPart,
    UnsupportedOperationError,
)
from a2a.utils import new_task
from a2a.utils.errors import ServerError


logger = logging.getLogger(__name__)
SUPPORTED_CONTENT_TYPES = ["text", "text/plain"]
A2A_FALLBACK_RESPONSE = "RoomAgent 已收到请求，当前返回仍为最简占位结果。"

_CURRENT_UPDATER: TaskUpdater | None = None


def get_current_updater() -> TaskUpdater:
    """Get the current TaskUpdater for the executing task, if available."""
    global _CURRENT_UPDATER
    if _CURRENT_UPDATER is None:
        raise RuntimeError("No current TaskUpdater available")
    return _CURRENT_UPDATER


def create_text_part(text: str) -> Part:
    """Helper function to create a text Part."""
    return Part(root=TextPart(text=text.strip()))


class RoomAgentExecutor(AgentExecutor):
    """Minimal A2A executor skeleton for the RoomAgent service."""

    async def execute(
        self,
        context: RequestContext,
        event_queue: EventQueue,
    ) -> None:
        task = context.current_task
        if not task:
            task = new_task(context.message)
            await event_queue.enqueue_event(task)

        updater = TaskUpdater(event_queue, task.id, task.context_id)
        global _CURRENT_UPDATER
        _CURRENT_UPDATER = updater

        user_input = context.get_user_input()
        logger.info("Received RoomAgent A2A request: %s", user_input)

        await self.invoke_roomagent_entrypoint(
            user_input=user_input,
            context_id=task.context_id,
            task_id=task.id,
            conversation_text=_build_conversation_text(
                task=task,
                current_message=context.message,
            ),
        )

    async def invoke_roomagent_entrypoint(
        self,
        *,
        user_input: str,
        context_id: str,
        task_id: str,
        conversation_text: str | None = None,
    ) -> dict[str, Any]:
        """Invoke the RoomAgent LangGraph entrypoint and return execution_result."""
        app = _compile_graph()

        updater = get_current_updater()
        await updater.start_work()

        final_state = await app.ainvoke(
            {
                "user_input": user_input,
                "conversation_text": (conversation_text or user_input).strip() or user_input,
                "metadata": {
                    "context_id": context_id,
                    "task_id": task_id,
                    "source": "a2a",
                },
            }
        )

        logger.info(
            "RoomAgent graph final state task_id=%s context_id=%s state=%s",
            task_id,
            context_id,
            json.dumps(final_state, ensure_ascii=False, default=str),
        )
        execution_result = final_state.get("execution_result", {})
        if isinstance(execution_result, dict):
            return execution_result
        return {}

    async def cancel(
        self,
        context: RequestContext,
        event_queue: EventQueue,
    ) -> None:
        _ = (context, event_queue)
        raise ServerError(error=UnsupportedOperationError())


def _build_conversation_text(
    *,
    task: Task | None,
    current_message: Message | None,
) -> str:
    history = list(task.history or []) if task and task.history else []
    if current_message and (not history or history[-1].message_id != current_message.message_id):
        history.append(current_message)

    transcript_lines: list[str] = []
    for message in history:
        text = _extract_text_from_message(message)
        if not text:
            continue
        transcript_lines.append(f"{_normalize_role(message.role)}: {text}")

    current_user_input = _extract_text_from_message(current_message)
    if not current_user_input and history:
        for message in reversed(history):
            if _normalize_role(message.role) == "user":
                current_user_input = _extract_text_from_message(message)
                if current_user_input:
                    break

    if not transcript_lines:
        return current_user_input

    if not current_user_input:
        return "Conversation transcript:\n" + "\n".join(transcript_lines)

    return (
        "Conversation transcript:\n"
        + "\n".join(transcript_lines)
        + "\n\nCurrent user input:\n"
        + current_user_input
    )


def _extract_text_from_message(message: Message | None) -> str:
    if message is None:
        return ""

    parts = []
    for part in message.parts:
        if isinstance(part.root, TextPart):
            text = part.root.text.strip()
            if text:
                parts.append(text)
    return "\n".join(parts)


def _normalize_role(role: object) -> str:
    value = getattr(role, "value", role)
    if value == "agent":
        return "assistant"
    if value == "user":
        return "user"
    return str(value)


def _compile_graph():
    from graph.entry import compile_graph

    return compile_graph()


def build_a2a_application(*, host: str, port: int) -> A2AStarletteApplication:
    """Create the RoomAgent A2A Starlette application."""
    request_handler = DefaultRequestHandler(
        agent_executor=RoomAgentExecutor(),
        task_store=InMemoryTaskStore(),
    )
    return A2AStarletteApplication(
        agent_card=build_agent_card(host=host, port=port),
        http_handler=request_handler,
    )


def build_agent_card(*, host: str, port: int) -> AgentCard:
    """Create a blank RoomAgent agent card placeholder."""
    return AgentCard(
        name="RoomAgent",
        description="Blank RoomAgent A2A service skeleton awaiting skill and executor wiring.",
        url=f"http://{host}:{port}/",
        version="0.1.0",
        default_input_modes=SUPPORTED_CONTENT_TYPES,
        default_output_modes=SUPPORTED_CONTENT_TYPES,
        capabilities=AgentCapabilities(streaming=False, push_notifications=False),
        skills=[
            AgentSkill(
                id="room_agent_placeholder",
                name="RoomAgent Placeholder",
                description="Placeholder skill card to be completed later.",
                tags=["placeholder"],
                input_modes=SUPPORTED_CONTENT_TYPES,
                output_modes=SUPPORTED_CONTENT_TYPES,
            )
        ],
    )



"""Internal state definitions for the agent-execution subgraph."""

from __future__ import annotations

from typing import Any, Literal, NotRequired, TypedDict
from langchain_core.tools import BaseTool


class FinalOutput(TypedDict, total=False):
    """Structured final output emitted by the agent loop."""

    message: str
    summary: str | None
    metadata: dict[str, Any] | None


class PlannerStep(TypedDict, total=False):
    """Single planner step emitted by the agent loop."""

    step_type: Literal["reason", "toolcall", "final_output"]
    is_done: bool
    reason_summary: str
    tool_name: str
    tool_args: dict[str, Any]
    final_output: FinalOutput


class ToolResult(TypedDict, total=False):
    """Internal record of a single tool execution."""

    step_index: int
    tool_name: str
    tool_args: dict[str, Any]
    observation: Any
    error: str
    args_summary: str
    result_summary: str
    error_summary: str


class AgentExecutionState(TypedDict, total=False):
    """Encapsulated state for the agent-execution loop.

    Field groups:

    - request context: copied in from the outer graph so the subgraph can plan
      against the original user request and the intent/tool-selection result
    - runtime loop state: mutated on each planner / tool-execution iteration
    - terminal state: written once the subgraph reaches final output or fails
    """

    # Original request context forwarded from the outer graph.
    user_input: str
    conversation_text: str
    # Intent recognized by the outer graph; used as planner context only.
    intent: dict[str, Any]
    # Tools pre-filtered by the outer tool_selection node. The subgraph should
    # only plan against this subset instead of re-selecting from the full tool list.
    selected_tools: list[dict[str, Any]]
    metadata: dict[str, Any]
    # Tools actually exposed to the planner/executor inside the subgraph.
    available_tools: list[dict[str, Any]]
    available_tool_instances: dict[str, BaseTool]

    # Mutable loop state.
    step_count: int
    step_limit: int
    current_step: PlannerStep
    step_history: list[dict[str, Any]]
    raw_model_outputs: list[str]
    tool_results: list[ToolResult]

    # Terminal / exit state.
    final_output: FinalOutput
    replan_used: bool
    loop_decision: NotRequired[str]
    terminal_error: dict[str, Any]
    outer_state_patch: dict[str, Any]



from __future__ import annotations

import pytest
from pydantic import ValidationError

from config.settings import HomeAssistantMCPSettings
from integrations.mcp_client import LangChainMCPClient, build_home_assistant_mcp_client


def test_build_home_assistant_mcp_client_returns_none_when_disabled() -> None:
    settings = HomeAssistantMCPSettings(enabled=False, base_url="http://ha.local:8123")

    client = build_home_assistant_mcp_client(settings)

    assert client is None


def test_build_home_assistant_mcp_client_uses_streamable_http_and_bearer_token() -> None:
    settings = HomeAssistantMCPSettings(
        enabled=True,
        server_name="home_assistant",
        transport="streamable_http",
        base_url="http://ha.local:8123",
        auth_token="secret-token",
    )

    client = build_home_assistant_mcp_client(settings)

    assert isinstance(client, LangChainMCPClient)
    connection = client._client.connections["home_assistant"]
    assert connection["transport"] == "streamable_http"
    assert connection["url"] == "http://ha.local:8123/api/mcp"
    assert connection["headers"] == {"Authorization": "Bearer secret-token"}


def test_build_home_assistant_mcp_client_normalizes_trailing_slash() -> None:
    settings = HomeAssistantMCPSettings(
        enabled=True,
        server_name="home_assistant",
        transport="streamable_http",
        base_url="http://ha.local:8123/",
    )

    client = build_home_assistant_mcp_client(settings)

    assert isinstance(client, LangChainMCPClient)
    connection = client._client.connections["home_assistant"]
    assert connection["url"] == "http://ha.local:8123/api/mcp"


def test_home_assistant_mcp_settings_reject_stdio_transport() -> None:
    with pytest.raises(ValidationError):
        HomeAssistantMCPSettings(
            enabled=True,
            transport="stdio",
            base_url="http://ha.local:8123",
        )



"""Direct-response node for low-cost conversational replies."""

from __future__ import annotations

from config.settings import LLMRole
from graph.state import RoomAgentGraphState
from app.a2a_server import create_text_part, get_current_updater


async def direct_response(state: RoomAgentGraphState) -> RoomAgentGraphState:
    """Generate a minimal natural-language reply for non-tool requests."""
    provider = _get_low_cost_provider()
    prompt_input = _get_prompt_input(state)
    reply = await provider.complete_text(_build_messages(prompt_input), json_mode=False)

    updater = get_current_updater()
    message = updater.new_agent_message(parts=[create_text_part(reply)])
    await updater.complete(message)

    return {
        "status": "completed",
        "next_action": "direct_response",
        "execution_result": {
            "type": "text",
            "message": reply.strip(),
            "intent": state.get("intent", {}),
        },
    }


def _get_low_cost_provider():
    from app.server import get_llm_provider_registry

    provider = get_llm_provider_registry().get(LLMRole.LOW_COST)
    if provider is None:
        raise RuntimeError(f"LLM provider is unavailable for role={LLMRole.LOW_COST.value}")
    return provider


def _build_messages(user_input: str) -> list[dict[str, str]]:
    return [
        {
            "role": "system",
            "content": (
                "你是 Room Agent 的直接回复节点。"
                "当前请求不需要工具调用。"
                "请直接给出一条简短、自然、礼貌的中文回复。"
                "不要编造工具执行，不要输出 JSON，不要输出多段长文。"
            ),
        },
        {
            "role": "user",
            "content": user_input,
        },
    ]


def _get_prompt_input(state: RoomAgentGraphState) -> str:
    return state.get("conversation_text", "").strip() or state.get("user_input", "").strip()



"""Intent recognition node for the Room Agent graph."""

from __future__ import annotations

from typing import Any

from config.settings import LLMRole
from graph.state import RoomAgentGraphState
from llm_json_parse import JsonParserWithRepair


INTENT_OUTPUT_SCHEMA = {
    "type": "object",
    "properties": {
        "intent_name": {"type": "string", "minLength": 1},
        "need_tool_call": {"type": "boolean"},
    },
    "required": ["intent_name", "need_tool_call"],
    "additionalProperties": False,
}


async def intent_recognition(state: RoomAgentGraphState) -> RoomAgentGraphState:
    """Classify whether the request needs tool execution."""
    provider = _get_low_cost_provider()
    prompt_input = _get_prompt_input(state)

    raw_output = await provider.complete_text(
        _build_messages(prompt_input),
        json_mode=True,
    )
    parsed = await JsonParserWithRepair(llm_provider=provider)(
        raw_output,
        schema=INTENT_OUTPUT_SCHEMA,
    )

    return {
        "intent": {
            "name": parsed["intent_name"],
        },
        "need_tool_call": parsed["need_tool_call"],
        "next_action": "tool_selection" if parsed["need_tool_call"] else "direct_response",
    }


def route_after_intent(state: RoomAgentGraphState) -> str:
    """Route to the next node after intent recognition."""
    return "tool_selection" if state.get("need_tool_call") else "direct_response"


def _get_low_cost_provider() -> Any:
    from app.server import get_llm_provider_registry

    provider = get_llm_provider_registry().get(LLMRole.LOW_COST)
    if provider is None:
        raise RuntimeError(f"LLM provider is unavailable for role={LLMRole.LOW_COST.value}")
    return provider


def _build_messages(user_input: str) -> list[dict[str, str]]:
    return [
        {
            "role": "system",
            "content": (
                "你是 Room Agent 的意图分析节点。"
                "你的任务只有一个：判断当前用户请求是否需要后续工具调用或执行能力。"
                "如果只是闲聊、问候、纯解释、普通文本问答，则 need_tool_call=false。"
                "如果请求需要外部工具、设备控制、查询系统或后续执行流程，则 need_tool_call=true。"
                "请仅输出 JSON，不要输出额外解释。"
            ),
        },
        {
            "role": "user",
            "content": (
                "请分析下面的用户输入，并输出 JSON。\n"
                "字段要求：intent_name(string), need_tool_call(boolean).\n"
                "下面是示例：\n"
                "示例1\n"
                "输入：你好\n"
                "输出：{\"intent_name\":\"chat\",\"need_tool_call\":false}\n"
                "示例2\n"
                "输入：你是谁\n"
                "输出：{\"intent_name\":\"chat\",\"need_tool_call\":false}\n"
                "示例3\n"
                "输入：帮我打开卧室的灯\n"
                "输出：{\"intent_name\":\"device_control\",\"need_tool_call\":true}\n"
                "示例4\n"
                "输入：查一下今天上海天气\n"
                "输出：{\"intent_name\":\"information_query\",\"need_tool_call\":true}\n"
                "示例5\n"
                "输入：解释一下为什么晚上开空调会口干\n"
                "输出：{\"intent_name\":\"explanation\",\"need_tool_call\":false}\n"
                "现在请只针对下面这个输入输出 JSON，不要附加解释。\n"
                f"用户输入：{user_input}"
            ),
        },
    ]


def _get_prompt_input(state: RoomAgentGraphState) -> str:
    return state.get("conversation_text", "").strip() or state.get("user_input", "").strip()



"""LLM-driven tool-selection node."""

from __future__ import annotations

import json
import sys
from typing import Any

from config.settings import LLMRole
from graph.state import RoomAgentGraphState
from langchain_core.tools import BaseTool
from llm_json_parse import JsonParserWithRepair


TOOL_SELECTION_OUTPUT_SCHEMA = {
    "type": "object",
    "properties": {
        "selected_tool_names": {
            "type": "array",
            "items": {"type": "string", "minLength": 1},
        },
        "comment": {"type": "string"},
    },
    "required": ["selected_tool_names", "comment"],
    "additionalProperties": False,
}


async def tool_selection(state: RoomAgentGraphState) -> RoomAgentGraphState:
    """Select up to three MCP tools for the current request."""
    provider = _get_low_cost_provider()
    prompt_input = _get_prompt_input(state)
    candidate_tools = await _describe_tools()

    if not candidate_tools:
        comment = "No MCP tools are available for this request."
        _log_zero_tool_selection(prompt_input, state.get("intent", {}), comment)
        return _build_result(candidate_tools, [], comment)

    raw_output = await provider.complete_text(
        _build_messages(
            prompt_input=prompt_input,
            intent=state.get("intent", {}),
            candidate_tools=candidate_tools,
        ),
        temperature=0,
        json_mode=True,
    )
    parsed = await JsonParserWithRepair(llm_provider=provider)(
        raw_output,
        schema=TOOL_SELECTION_OUTPUT_SCHEMA,
    )

    selected_tools = _select_tools(
        candidate_tools,
        parsed.get("selected_tool_names", []),
    )
    comment = parsed.get("comment", "").strip() or "No tool-selection comment provided."

    if not selected_tools:
        _log_zero_tool_selection(prompt_input, state.get("intent", {}), comment)

    return _build_result(candidate_tools, selected_tools, comment)


def _get_low_cost_provider() -> Any:
    from app.server import get_llm_provider_registry

    provider = get_llm_provider_registry().get(LLMRole.LOW_COST)
    if provider is None:
        raise RuntimeError(f"LLM provider is unavailable for role={LLMRole.LOW_COST.value}")
    return provider


def _get_mcp_client():
    from app.server import get_mcp_client

    return get_mcp_client()


async def _describe_tools() -> list[dict[str, Any]]:
    client = _get_mcp_client()
    if client is None:
        return []

    tools = await client.get_tools()
    return [
        {
            "name": tool.name,
            "description": tool.description or "",
            "args_schema": _extract_tool_schema(tool),
        }
        for tool in tools
    ]


def _build_messages(
    *,
    prompt_input: str,
    intent: dict[str, Any],
    candidate_tools: list[dict[str, Any]],
) -> list[dict[str, str]]:
    return [
        {
            "role": "system",
            "content": (
                "你是 Room Agent 的工具选择节点。"
                "你的职责只有一个：从候选 MCP 工具中挑选最适合当前请求的 0 到 3 个工具。如果用户提到了多个意图，你可以选择多个。按执行顺序或者相关程度排序。"
                "如果没有任何合适工具，返回空数组。"
                "只输出 JSON，不要输出额外解释。"
            ),
        },
        {
            "role": "user",
            "content": json.dumps(
                {
                    "task": (
                        "请基于用户输入、已识别意图和候选工具，选择最合适的 0 到 3 个工具。"
                        "输出 JSON，字段为 selected_tool_names(string[]) 和 comment(string)。"
                    ),
                    "user_input": prompt_input,
                    "intent": intent,
                    "candidate_tools": candidate_tools,
                },
                ensure_ascii=False,
            ),
        },
    ]


def _normalize_tool_name(name: str) -> str:
    return name.strip().lower().replace(" ", "_").replace("-", "_")


def _select_tools(
    candidate_tools: list[dict[str, Any]],
    selected_tool_names: list[str],
) -> list[dict[str, Any]]:
    tool_by_name = {_normalize_tool_name(tool["name"]): tool for tool in candidate_tools}
    selected_tools: list[dict[str, Any]] = []
    seen_names: set[str] = set()

    for tool_name in selected_tool_names:
        _normalized_selected_name = _normalize_tool_name(tool_name)

        if _normalized_selected_name in seen_names:
            continue
        tool = tool_by_name.get(_normalized_selected_name)
        if tool is None:
            continue
        selected_tools.append(tool)
        seen_names.add(_normalized_selected_name)

    # 提示词防止模型发狂, 实际上我们不写死上限
    return selected_tools


def _build_result(
    candidate_tools: list[dict[str, Any]],
    selected_tools: list[dict[str, Any]],
    comment: str,
) -> RoomAgentGraphState:
    return {
        "status": "completed",
        "next_action": "tool_selection",
        "candidate_tools": candidate_tools,
        "selected_tools": selected_tools,
        "execution_result": {
            "type": "tool_selection",
            "comment": comment,
            "selected_count": len(selected_tools),
        },
    }


def _log_zero_tool_selection(
    prompt_input: str,
    intent: dict[str, Any],
    comment: str,
) -> None:
    print(
        "[tool_selection] no tools selected "
        f"intent={json.dumps(intent, ensure_ascii=False)} "
        f"comment={comment} "
        f"user_input={prompt_input}",
        file=sys.stderr,
    )


def _get_prompt_input(state: RoomAgentGraphState) -> str:
    return state.get("conversation_text", "").strip() or state.get("user_input", "").strip()


def _extract_tool_schema(tool: BaseTool) -> dict[str, Any]:
    get_input_schema = getattr(tool, "get_input_schema", None)
    if callable(get_input_schema):
        schema_model = get_input_schema()
        model_json_schema = getattr(schema_model, "model_json_schema", None)
        if callable(model_json_schema):
            schema = model_json_schema()
            return schema if isinstance(schema, dict) else {}

    args_schema = getattr(tool, "args", {}) or {}
    if args_schema:
        return {
            "type": "object",
            "properties": args_schema,
        }

    return {}



"""Gateway request stubs for RoomAgent runtime integration."""

from __future__ import annotations

from config.settings import AgentSettings

# TODO 对齐和后端的协议之后, 填入真实实现
class GatewayClient:
    """Placeholder HTTP client for backend gateway integration."""

    async def register_agent(self, agent: AgentSettings) -> None:
        """Send the startup registration request to the backend gateway."""
        return None

    async def send_heartbeat(self, agent: AgentSettings) -> None:
        """Send a heartbeat request to the backend gateway."""
        return None



#!/usr/bin/env python3
"""Minimal A2A debug client for local service testing."""

from __future__ import annotations

import argparse
import asyncio
import json
from uuid import uuid4

import httpx
from a2a.client import A2ACardResolver, ClientConfig, ClientFactory
from a2a.client.helpers import create_text_message_object
from a2a.types import Message, MessageSendConfiguration, Task, TaskQueryParams


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Debug an A2A service by resolving its agent card and sending JSON-RPC requests.",
    )
    parser.add_argument(
        "--url",
        default="http://127.0.0.1:10000",
        help="Base URL of the A2A service. Defaults to http://127.0.0.1:10000.",
    )
    parser.add_argument(
        "--card-path",
        default=None,
        help="Optional relative path to the agent card. Defaults to the SDK well-known path.",
    )
    parser.add_argument(
        "--header",
        action="append",
        default=[],
        help="Extra HTTP header in 'Key: Value' form. Repeatable.",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        default=30.0,
        help="HTTP timeout in seconds. Use 0 for no timeout.",
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    subparsers.add_parser("card", help="Fetch and print the resolved agent card.")

    send_parser = subparsers.add_parser("send", help="Send one user message to the agent.")
    send_parser.add_argument("message", help="User message text.")
    send_parser.add_argument("--task-id", default=None, help="Optional existing task ID.")
    send_parser.add_argument("--context-id", default=None, help="Optional existing context ID.")
    send_parser.add_argument(
        "--history-length",
        type=int,
        default=None,
        help="Optional historyLength to attach to message/send.",
    )

    get_task_parser = subparsers.add_parser("get-task", help="Fetch an existing task.")
    get_task_parser.add_argument("task_id", help="Task ID to fetch.")
    get_task_parser.add_argument(
        "--history-length",
        type=int,
        default=None,
        help="Optional historyLength for tasks/get.",
    )

    return parser.parse_args()


def parse_headers(raw_headers: list[str]) -> dict[str, str]:
    headers: dict[str, str] = {}
    for item in raw_headers:
        if ":" not in item:
            raise SystemExit(f"Invalid --header value: {item!r}. Expected 'Key: Value'.")
        key, value = item.split(":", 1)
        headers[key.strip()] = value.strip()
    return headers


async def resolve_agent_card(
    *,
    httpx_client: httpx.AsyncClient,
    url: str,
    card_path: str | None,
):
    resolver = A2ACardResolver(httpx_client=httpx_client, base_url=url)
    return await resolver.get_agent_card(relative_card_path=card_path)


async def run() -> int:
    args = parse_args()
    headers = parse_headers(args.header)
    timeout = None if args.timeout == 0 else args.timeout

    async with httpx.AsyncClient(headers=headers, timeout=timeout) as httpx_client:
        agent_card = await resolve_agent_card(
            httpx_client=httpx_client,
            url=args.url,
            card_path=args.card_path,
        )

        if args.command == "card":
            print(
                json.dumps(
                    agent_card.model_dump(mode="json", by_alias=True), ensure_ascii=False, indent=2
                )
            )
            return 0

        client = await ClientFactory.connect(
            agent_card,
            client_config=ClientConfig(
                streaming=False,
                httpx_client=httpx_client,
            ),
        )

        if args.command == "send":
            message = create_text_message_object(content=args.message)
            if args.task_id:
                message.task_id = args.task_id
            if args.context_id:
                message.context_id = args.context_id

            configuration = None
            if args.history_length is not None:
                configuration = MessageSendConfiguration(history_length=args.history_length)

            response = await _send_message(
                client,
                message=message,
                configuration=configuration,
            )
            print(
                json.dumps(
                    response.model_dump(mode="json", by_alias=True, exclude_none=True),
                    ensure_ascii=False,
                    indent=2,
                )
            )
            return 0

        if args.command == "get-task":
            response = await client.get_task(
                TaskQueryParams(
                    id=args.task_id,
                    history_length=args.history_length,
                )
            )
            print(
                json.dumps(
                    response.model_dump(mode="json", by_alias=True, exclude_none=True),
                    ensure_ascii=False,
                    indent=2,
                )
            )
            return 0

    raise SystemExit(f"Unsupported command: {args.command}")


async def _send_message(
    client,
    *,
    message: Message,
    configuration: MessageSendConfiguration | None,
) -> Task | Message:
    async for event in client.send_message(
        message,
        configuration=configuration,
    ):
        if isinstance(event, tuple):
            task, _ = event
            return task
        return event
    raise RuntimeError("A2A client returned no response event.")


def cli():
    raise SystemExit(asyncio.run(run()))


if __name__ == "__main__":
    cli()



"""Heuristic intent classification helpers."""

from __future__ import annotations

from typing import Any


def heuristic_classify(user_input: str) -> dict[str, Any]:
    normalized = user_input.strip()
    lowered = normalized.lower()

    simple_chat_pairs = [
        ("你好", "你好，我在。"),
        ("hello", "你好，我在。"),
        ("hi", "你好，我在。"),
        ("谢谢", "不客气。"),
        ("你是谁", "我是 room-agent，目前负责房间任务路由。"),
    ]
    for keyword, response in simple_chat_pairs:
        if keyword in lowered or keyword in normalized:
            return {
                "intent_type": "simple_chat",
                "response": response,
            }

    device_verbs = ("打开", "关闭", "调高", "调低", "调亮", "调暗", "设置", "启动", "停止")
    device_nouns = ("灯", "空调", "窗帘", "风扇", "电视", "humidifier", "light")
    if any(verb in normalized for verb in device_verbs) and any(
        noun in lowered or noun in normalized for noun in device_nouns
    ):
        return {
            "intent_type": "task_request",
            "task_info": {
                "executor_type": "device",
                "task_name": "device_control",
                "parameters": {
                    "user_intent": normalized,
                    "context": {
                        "query": normalized,
                        "location": "",
                    },
                },
            },
        }

    mcp_keywords = ("天气", "搜索", "查询", "查一下", "新闻", "时间", "几点", "百科")
    if any(keyword in normalized for keyword in mcp_keywords):
        return {
            "intent_type": "task_request",
            "task_info": {
                "executor_type": "mcp",
                "task_name": "mcp_request",
                "parameters": {
                    "user_intent": normalized,
                    "context": {
                        "query": normalized,
                        "location": "",
                    },
                },
            },
        }

    return {
        "intent_type": "simple_chat",
        "response": "我先理解成普通对话。你也可以直接告诉我想查什么，或者要控制哪个设备。",
    }



#!/usr/bin/env python3
"""测试Room Agent心跳监听"""
import paho.mqtt.client as mqtt
import json
from datetime import datetime

# MQTT配置
BROKER = "120.78.228.69"
PORT = 1884
ROOM_ID = "livingroom"
AGENT_ID = "room-agent-livingroom"

# Heartbeat topic
HEARTBEAT_TOPIC = f"room/{ROOM_ID}/agent/{AGENT_ID}/heartbeat"
STATE_TOPIC = f"room/{ROOM_ID}/agent/{AGENT_ID}/state"

def on_connect(client, userdata, flags, rc):
    """连接回调"""
    if rc == 0:
        print("✓ 已连接到MQTT Broker")
        print(f"  Broker: {BROKER}:{PORT}")
        print()

        # 订阅heartbeat和state
        client.subscribe(HEARTBEAT_TOPIC)
        client.subscribe(STATE_TOPIC)
        print(f"✓ 已订阅topic:")
        print(f"  - {HEARTBEAT_TOPIC}")
        print(f"  - {STATE_TOPIC}")
        print()
        print("=" * 60)
        print("开始监听消息...")
        print("=" * 60)
    else:
        print(f"✗ 连接失败，错误码: {rc}")

def on_message(client, userdata, msg):
    """消息回调"""
    timestamp = datetime.now().strftime("%H:%M:%S")
    topic = msg.topic
    payload = msg.payload.decode("utf-8")

    print(f"\n[{timestamp}] 收到消息:")
    print(f"Topic: {topic}")

    try:
        # 尝试解析JSON
        data = json.loads(payload)
        print(f"内容: {json.dumps(data, indent=2, ensure_ascii=False)}")
    except json.JSONDecodeError:
        # 不是JSON，直接显示
        print(f"内容: {payload}")

    print("-" * 60)

def main():
    """主函数"""
    print("=" * 60)
    print("Room Agent 心跳监听工具")
    print("=" * 60)
    print(f"Room ID: {ROOM_ID}")
    print(f"Agent ID: {AGENT_ID}")
    print()

    # 创建MQTT客户端
    client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)
    client.on_connect = on_connect
    client.on_message = on_message

    # 连接到Broker
    print("正在连接到MQTT Broker...")
    try:
        client.connect(BROKER, PORT, 60)
    except Exception as e:
        print(f"✗ 连接失败: {e}")
        return

    # 保持连接
    client.loop_forever()

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("\n\n停止监听")
